###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Command to apply a pre-defined configuration to PMM"""
from __future__ import absolute_import

from copy import deepcopy

try:
    from rdmc_helper import (
        LOGGER,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoChangesFoundOrMadeError,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        LOGGER,
        NoChangesFoundOrMadeError,
        NoContentsFoundForOperationError,
    )

from .lib.DisplayHelpers import DisplayHelpers
from .lib.RestHelpers import RestHelpers


class ApplyPmemConfigCommand:
    """
    Command to apply a pre-defined configuration to PMM
    """

    # Available configIDs
    config_ids = [
        {
            "name": "MemoryMode",
            "description": "Configure all the PMMs to 100% Memory Mode.",
            "totalPmemPercentage": 0,
            "pmemIntereleaved": False,
        },
        {
            "name": "PmemInterleaved",
            "description": "Configure all PMMs to 100% Persistent. "
            "Interleave the Persistent memory regions within each processor.",
            "totalPmemPercentage": 100,
            "pmemIntereleaved": True,
        },
        {
            "name": "PmemNotInterleaved",
            "description": "Configure all PMMs to 100% Persistent. "
            "The Persistent memory regions are not interleaved.",
            "totalPmemPercentage": 100,
            "pmemIntereleaved": False,
        },
    ]

    def __init__(self):
        self.ident = {
            "name": "applypmmconfig",
            "usage": None,
            "description": "Applies a pre-defined configuration to PMM\n"
            "\texample: applypmmconfig --pmmconfig MemoryMode\n",
            "summary": "Applies a pre-defined configuration to PMM.",
            "aliases": [],
            "auxcommands": [
                "ShowPmemPendingConfigCommand",
                "ClearPendingConfigCommand",
            ],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

        self._display_helpers = DisplayHelpers()

    def definearguments(self, customparser):
        """
        Wrapper function for new command main function
        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-C",
            "--pmmconfig",
            action="store",
            type=str,
            dest="config",
            help="Specify one of the pre-defined configIDs to apply" " to all Persistent Memory Modules.",
            default=None,
        )

        customparser.add_argument(
            "-L",
            "--list",
            action="store_true",
            dest="list",
            help="Display a list of available pre-defined configIDs" " along with a brief description.",
            default=False,
        )

        customparser.add_argument(
            "-f",
            "--force",
            action="store_true",
            dest="force",
            help="Allow the user to force the configuration " "by automatically accepting any prompts.",
            default=False,
        )

    def run(self, line, help_disp=False):
        """
        Wrapper function for new command main function
        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        LOGGER.info("PMM Apply Pre-Defined Configuration: %s", self.ident["name"])
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineError("Failed to parse options")
        if args:
            self.validate_args(options)
        self.validate_options(options)
        self.apply_pmm_config(options)

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    @staticmethod
    def validate_args(options):
        """
        Produces relevant error messages when unwanted extra arguments are specified with flags.
        """
        if options.config:
            raise InvalidCommandLineError("'config | -C' expects a single argument")
        else:
            raise InvalidCommandLineError("Chosen flag doesn't expect additional arguments")

    def validate_options(self, options):
        """
        Validate whether options specified by user are valid.
        :param options: options specified by the user.
        :type options: instance of OptionParser class.
        """
        # Set of valid configIDs
        valid_config_ids = {config_id.get("name").lower() for config_id in self.config_ids}

        if not options.config and not options.list and not options.force:
            raise InvalidCommandLineError("No flag specified.\n\nUsage: " + self.ident["usage"])
        if not options.config and options.force:
            raise InvalidCommandLineError("'--force | -f' flag mandatorily requires the " "'--pmmconfig | -C' flag.")
        if options.config and options.list:
            raise InvalidCommandLineError("Only one of '--pmmconfig | -C' or '--list | -L' " "may be specified")
        # Check whether the user entered configID is valid
        if options.config and not options.config.lower() in valid_config_ids:
            raise InvalidCommandLineError(
                "Invalid configID.\nTo view a list of pre-defined " "configIDs, use 'applypmmconfig --list'"
            )

    def apply_pmm_config(self, options):
        """
        Calls relevant functions based on the flags specified by the user.
        :param options: options specified by the user.
        :type options: instance of OptionParser class.
        """
        if options.list:
            self.show_predefined_config_options()
        elif options.config:
            # Raise exception if server is in POST
            if RestHelpers(rdmcObject=self.rdmc).in_post():
                raise NoContentsFoundForOperationError(
                    "Unable to retrieve resources - " "server might be in POST or powered off"
                )
            self.apply_predefined_config(options)

    def show_predefined_config_options(self):
        """
        Display the available pre-defined configIDs that the user can choose from to
        apply to their Persistent memory Modules.
        """
        self.rdmc.ui.printer("\nAvailable Configurations:\n\n")
        for config_id in ApplyPmemConfigCommand.config_ids:
            self.rdmc.ui.printer(config_id.get("name") + "\n")
            self.rdmc.ui.printer("\t" + config_id.get("description") + "\n")
        self.rdmc.ui.printer("\n")

    def warn_existing_chunks_and_tasks(self, memory_chunk_tasks, memory_chunks):
        """
        Checks for existing Memory Chunks and Pending Configuration Task resources on
        a server where a user is trying to apply a pre-defined configuration
        :param memory_chunk_tasks: Pending Configuration Tasks
        :type memory_chunk_tasks: list
        :memory_chunks: Memory Chunks in the existing configuration
        :type memory_chunks: list
        :returns: None
        """
        # If Memory Chunks exist, display Existing configuration warning
        if memory_chunks:
            self.rdmc.ui.printer(
                "Warning: Existing configuration found. Proceeding with applying a new "
                "configuration will result in overwriting the current configuration and "
                "cause data loss.\n"
            )
        # If Pending Configuration Tasks exist, display warning
        if memory_chunk_tasks:
            self.rdmc.ui.printer(
                "Warning: Pending configuration tasks found. Proceeding with applying "
                "a new configuration will result in overwriting the pending "
                "configuration tasks.\n"
            )
        # Raise a NoChangesFoundOrMade exception when either of the above conditions exist
        if memory_chunks or memory_chunk_tasks:
            # Line feed for proper formatting
            raise NoChangesFoundOrMadeError(
                "\nFound one or more of Existing Configuration or "
                "Pending Configuration Tasks. Please use the "
                "'--force | -f' flag with the same command to "
                "approve these changes."
            )
        return None

    def delete_existing_chunks_and_tasks(self, memory_chunk_tasks, memory_chunks):
        """
        Delete existing Memory Chunks and Pending Configuration Tasks
        :param memory_chunk_tasks: Pending Configuration Tasks
        :type memory_chunk_tasks: list
        :memory_chunks: Memory Chunks in the existing configuration
        :type memory_chunks: list
        :returns: None
        """
        # Delete any pending configuration tasks
        if memory_chunk_tasks:
            self.auxcommands["clearpmmpendingconfig"].delete_tasks(memory_chunk_tasks)
        # Delete any existing configuration
        if memory_chunks:
            for chunk in memory_chunks:
                data_id = chunk.get("@odata.id")
                resp = RestHelpers(rdmcObject=self.rdmc).delete_resource(data_id)
                if not resp:
                    raise NoChangesFoundOrMadeError("Error occured while deleting " "existing configuration")
        return None

    @staticmethod
    def get_post_data(config_data, interleavable_memory_sets):
        """
        Create post body based on config data
        :param config_data: An entry of self.config_iD's list
        :type config_data: dict
        :param interleavable_memory_sets: List of interleavable sets
        :type interleavable_memory_sets: List
        :returns: Returns post data
        """

        body = {
            "AddressRangeType": "PMEM",
            "InterleaveSets": [],
            "Oem": {"Hpe": {"MemoryChunkSizePercentage": config_data["totalPmemPercentage"]}},
        }
        # Get the list of interleave sets based on the configuration
        if config_data["pmemIntereleaved"] or config_data["totalPmemPercentage"] == 0:
            # If pmem regions are interleaved or if it MemoryMode, choose the list with maximum entries.
            interleave_sets = [max(interleavable_memory_sets, key=lambda x: len(x["MemorySet"]))]
        else:
            # If pmem regions are NOT interleaved, choose all the lists with exactly one entry.
            interleave_sets = [il_set for il_set in interleavable_memory_sets if len(il_set["MemorySet"]) == 1]

        # Using in-place update to change the interleave sets format.
        # Replace 'MemorySet' with 'Memory' for each MemorySet in interleave_sets.
        # Get '@odata.id' for each memory region from original interleave_sets and add this to new structure.
        for index, interleavableset in enumerate(interleave_sets):
            interleave_sets[index] = [
                {"Memory": {"@odata.id": str(memory_region["@odata.id"])}}
                for memory_region in interleavableset["MemorySet"]
            ]

        # Create post body for each interleave sets from post body template.
        post_data = []
        for interleaveset in interleave_sets:
            current_body = deepcopy(body)
            current_body["InterleaveSets"] = interleaveset
            post_data.append(current_body)

        return post_data

    def apply_predefined_config(self, options):
        """
        Apply the selected pre-defined configuration to Persistent Memory Modules.
        :param options: option specified by the user.
        :returns: None
        """
        # Retrieve Memory Chunks and Task Resources from server
        (task_members, domain_members, memory_chunks) = RestHelpers(
            rdmcObject=self.rdmc
        ).retrieve_task_members_and_mem_domains()

        # Filter Task Resources to include only Pending Configuration Tasks
        memory_chunk_tasks = RestHelpers(rdmcObject=self.rdmc).filter_task_members(task_members)

        if options.force:
            self.delete_existing_chunks_and_tasks(memory_chunk_tasks, memory_chunks)
        else:
            self.warn_existing_chunks_and_tasks(memory_chunk_tasks, memory_chunks)

        if not domain_members:
            raise NoContentsFoundForOperationError("Failed to retrive Memory Domain Resources")

        # Get the user specified configID
        config_data = next(
            (config_id for config_id in self.config_ids if config_id.get("name").lower() == options.config.lower()),
            None,
        )

        for proc in domain_members:
            path = proc["MemoryChunks"].get("@odata.id")
            data = self.get_post_data(config_data, proc["InterleavableMemorySets"])
            for body in data:
                resp = RestHelpers(rdmcObject=self.rdmc).post_resource(path, body)
                if resp is None:
                    raise NoChangesFoundOrMadeError("Error occured while applying configuration")

        # display warning
        self.rdmc.ui.printer("Configuration changes require reboot to take effect.\n")

        # display pending configuration
        self.auxcommands["showpmmpendingconfig"].show_pending_config(type("MyOptions", (object,), dict(json=False)))



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Log Operations Command for rdmc """

import ctypes
import datetime
import itertools
import json
import os
import platform
import shlex
import string
import subprocess
import sys
import tempfile
import time

from six.moves import queue

import redfish.hpilo.risblobstore2 as risblobstore2
from redfish.rest.connections import SecurityStateError
from redfish.ris.utils import filter_output

try:
    from rdmc_helper import (
        LOGGER,
        UI,
        IncompatibleiLOVersionError,
        InvalidCListFileError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileInputError,
        InvalidKeyError,
        InvalidMSCfileInputError,
        MultipleServerConfigError,
        NoContentsFoundForOperationError,
        PartitionMoutingError,
        ReturnCodes,
        UnabletoFindDriveError,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        LOGGER,
        UI,
        IncompatibleiLOVersionError,
        InvalidCListFileError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileInputError,
        InvalidKeyError,
        InvalidMSCfileInputError,
        MultipleServerConfigError,
        NoContentsFoundForOperationError,
        PartitionMoutingError,
        ReturnCodes,
        UnabletoFindDriveError,
    )

if os.name == "nt":
    import win32api
elif sys.platform != "darwin" and "VMkernel" not in platform.uname():
    import pyudev


class ServerlogsCommand:
    """Download logs from the server that is currently logged in"""

    def __init__(self):
        self.ident = {
            "name": "serverlogs",
            "usage": None,
            "description": "Download the AHS"
            " logs from the logged in server.\n\texample: serverlogs "
            "--selectlog=AHS \n\n\tClear the AHS logs "
            "from the logged in server.\n\texample: serverlogs "
            "--selectlog=AHS --clearlog\n\n\tDownload the IEL"
            " logs from the logged in server.\n\texample: serverlogs "
            "--selectlog=IEL -f IELlog.txt\n\n\tClear the IEL logs "
            "from the logged in server.\n\texample: serverlogs "
            "--selectlog=IEL --clearlog\n\n\tDownload the IML"
            " logs from the logged in server.\n\texample: serverlogs "
            "--selectlog=IML -f IMLlog.txt\n\n\tClear the IML logs "
            "from the logged in server.\n\texample: serverlogs "
            "--selectlog=IML --clearlog\n\n\t(IML LOGS ONLY FEATURE)"
            "\n\tInsert entry in the IML logs from the logged in "
            'server.\n\texample: serverlogs --selectlog=IML -m "Text'
            ' message for maintenance"\n\n\tDownload the iLO Security'
            " logs from the logged in server.\n\texample: serverlogs "
            "--selectlog=SL -f SLlog.txt\n\n\tClear the iLO Security logs "
            "from the logged in server.\n\texample: serverlogs "
            "--selectlog=SL --clearlog\n\n\tDownload logs from multiple servers"
            "\n\texample: serverlogs --mpfile mpfilename.txt -o output"
            "directorypath --mplog=IEL,IML\n\n\tNote: multiple server file "
            "format (1 server per new line)\n\t--url <iLO url/hostname> "
            "-u admin -p password\n\t--url <iLO url/hostname> -u admin -"
            "p password\n\t--url <iLO url/hostname> -u admin -p password"
            "\n\n\tInsert customised string "
            "if required for AHS\n\texample: serverlogs --selectlog="
            'AHS --customiseAHS "from=2014-03-01&&to=2014'
            '-03-30"\n\n\t(AHS LOGS ONLY FEATURE)\n\tInsert the location/'
            "path of directory where AHS log needs to be saved."
            " \n\texample: serverlogs --selectlog=AHS "
            "--directorypath=C:\\Python38\\DataFiles\n\n\tRepair IML log."
            "\n\texample: serverlogs --selectlog=IML --repair IMLlogID",
            "summary": "Download and perform log operations.",
            "aliases": ["logservices"],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()
        self.dontunmount = None
        self.queue = queue.Queue()
        self.abspath = None
        self.lib = None

    def run(self, line, help_disp=False):
        """Main serverlogs function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if not getattr(options, "sessionid", False):
            self.serverlogsvalidation(options)

        if options.mpfilename:
            self.rdmc.ui.printer("Downloading logs for multiple servers...\n")
            return self.gotompfunc(options)

        self.serverlogsworkerfunction(options)

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def serverlogsworkerfunction(self, options):
        """ "Main worker function outlining the process

        :param options: command line options
        :type options: list.
        """
        if not options.service:
            raise InvalidCommandLineError("Please select a log type using the --selectlog option.")

        if options.service.lower() == "iml":
            path = self.returnimlpath(options=options)
        elif options.service.lower() == "iel":
            path = self.returnielpath(options=options)
        elif options.service.lower() == "sl":
            path = self.returnslpath(options=options)
        elif options.service.lower() == "ahs" and options.filter:
            raise InvalidCommandLineError("Cannot filter AHS logs.")
        elif (
            options.service.lower() == "ahs"
            and (not self.rdmc.app.typepath.url or self.rdmc.app.typepath.url.startswith("blobstore"))
            and not options.clearlog
        ):
            if options.customiseAHS is not None:
                current_date = str(datetime.datetime.now()).split()[0]
                date_check = options.customiseAHS
                from_date = date_check.split("&")[0].split("=")[1]
                to_date = date_check.split("to")[-1].split("=")[-1]
                if to_date < from_date or from_date > current_date:
                    raise InvalidCommandLineError("Please provide valid date to customiseAHS")
            self.downloadahslocally(options=options)
            return
        elif options.service.lower() == "ahs":
            path = self.returnahspath(options)
        else:
            raise InvalidCommandLineError("Log opted does not exist!")

        data = None

        if options.clearlog:
            self.clearlog(path)
        elif options.mainmes:
            self.addmaintenancelogentry(options, path=path)
        elif options.repiml:
            self.repairlogentry(options, path=path)
        else:
            data = self.downloaddata(path=path, options=options)

        self.savedata(options=options, data=data)

    def gotompfunc(self, options):
        """ "Function to download logs from multiple servers concurrently

        :param options: command line options
        :type options: list.
        """
        if options.mpfilename:
            mfile = options.mpfilename
            outputdir = None

            if options.outdirectory:
                outputdir = options.outdirectory

            if self.runmpfunc(mpfile=mfile, outputdir=outputdir, options=options):
                return ReturnCodes.SUCCESS
            else:
                raise MultipleServerConfigError("One or more servers failed to download logs.")

    def runmpfunc(self, mpfile=None, outputdir=None, options=None):
        """Main worker function for multi file command

        :param mpfile: configuration file
        :type mpfile: string.
        :param outputdir: custom output directory
        :type outputdir: string.
        """
        self.logoutobj.run("")
        LOGGER.info("Validating input server collection file.")
        data = self.validatempfile(mpfile=mpfile, options=options)

        if not data:
            return False

        processes = []
        finalreturncode = True
        outputform = "%Y-%m-%d-%H-%M-%S"

        if outputdir:
            if outputdir.endswith(('"', "'")) and outputdir.startswith(('"', "'")):
                outputdir = outputdir[1:-1]

            if not os.path.isdir(outputdir):
                raise InvalidCommandLineError("The given output folder path does not exist.")

            dirpath = outputdir
        else:
            dirpath = os.getcwd()

        dirname = "%s_%s" % (datetime.datetime.now().strftime(outputform), "MSClogs")
        createdir = os.path.join(dirpath, dirname)
        os.mkdir(createdir)

        oofile = open(os.path.join(createdir, "CompleteOutputfile.txt"), "w+")
        self.rdmc.ui.printer("Creating multiple processes to load configuration " "concurrently to all servers...\n")

        while True:
            if not self.queue.empty():
                line = self.queue.get()
            else:
                break

            finput = "\n" + "Output for " + line[line.index("--url") + 1] + ": \n\n"
            urlvar = line[line.index("--url") + 1]
            urlfilename = urlvar.split("//")[-1]
            line[line.index("-f") + 1] = str(line[line.index("-f") + 1]) + urlfilename

            if "python" in os.path.basename(sys.executable.lower()):
                # If we are running from source we have to add the python file to the command
                listargforsubprocess = [sys.executable, sys.argv[0]] + line
            else:
                listargforsubprocess = [sys.executable] + line

            if os.name != "nt":
                listargforsubprocess = " ".join(listargforsubprocess)

            logfile = open(os.path.join(createdir, urlfilename + ".txt"), "w+")
            pinput = subprocess.Popen(listargforsubprocess, shell=True, stdout=logfile, stderr=logfile)

            processes.append((pinput, finput, urlvar, logfile))

        for pinput, finput, urlvar, logfile in processes:
            pinput.wait()
            returncode = pinput.returncode
            finalreturncode = finalreturncode and not returncode

            logfile.close()
            logfile = open(os.path.join(createdir, urlvar + ".txt"), "r+")
            oofile.write(finput + str(logfile.read()))
            oofile.write("-x+x-" * 16)
            logfile.close()

            if returncode == 0:
                self.rdmc.ui.printer("Loading Configuration for {} : SUCCESS\n".format(urlvar))
            else:
                self.rdmc.ui.error("Loading Configuration for {} : FAILED\n".format(urlvar))
                self.rdmc.ui.error(
                    "ILOREST return code : {}.\nFor more details please check {}"
                    ".txt under {} directory.\n".format(returncode, urlvar, createdir)
                )

        oofile.close()

        if finalreturncode:
            self.rdmc.ui.printer("All servers have been successfully configured.\n")

        return finalreturncode

    def validatempfile(self, mpfile=None, options=None):
        """Validate temporary file

        :param mpfile: configuration file
        :type mpfile: string.
        :param lfile: custom file name
        :type lfile: string.
        """
        self.rdmc.ui.printer("Checking given server information...\n")

        if not mpfile:
            return False
        elif mpfile.startswith(('"', "'")) and mpfile[0] == mpfile[-1]:
            mpfile = mpfile[1:-1]

        if not os.path.isfile(mpfile) or not options.mplog:
            raise InvalidFileInputError(
                "File '%s' doesn't exist, please " "create file by running save command." % mpfile
            )

        logs = self.checkmplog(options)

        try:
            with open(mpfile, "r") as myfile:
                data = list()
                cmdtorun = ["serverlogs"]
                globalargs = ["-v", "--nocache"]

                while True:
                    line = myfile.readline()
                    if not line:
                        break

                    for logval in logs:
                        cmdargs = ["--selectlog=" + str(logval), "-f", str(logval)]
                        if line.endswith(os.linesep):
                            line.rstrip(os.linesep)

                        args = shlex.split(line, posix=False)

                        if len(args) < 5:
                            self.rdmc.ui.error("Incomplete data in input file: {}\n".format(line))
                            raise InvalidMSCfileInputError("Please verify the " "contents of the %s file" % mpfile)
                        else:
                            linelist = globalargs + cmdtorun + args + cmdargs
                            line = str(line).replace("\n", "")
                            self.queue.put(linelist)
                            data.append(linelist)
        except Exception as excp:
            LOGGER.info("%s", str(excp))
            raise excp

        if data:
            return data

        return False

    def checkmplog(self, options):
        """Function to validate mplogs options

        :param options: command line options
        :type options: list.
        """
        if options.mplog:
            logs = str(options.mplog)
            if "," in logs:
                logs = logs.split(",")
                return logs
            if logs in ("all", "IEL", "IML", "AHS"):
                if logs == "all":
                    logs = ["IEL", "IML", "AHS"]
                else:
                    logs = [logs]
                return logs
        raise InvalidCommandLineError("Error in mplogs options")

    def addmaintenancelogentry(self, options, path=None):
        """Worker function to add maintenance log

        :param options: command line options
        :type options: list.
        :param path: path to post maintenance log
        :type path: str
        """
        LOGGER.info("Adding maintenance logs")
        if options.mainmes is None:
            raise InvalidCommandLineError("")

        if options.service != "IML":
            raise InvalidCommandLineError("Log opted cannot make maintenance entries!")

        message = options.mainmes

        if message.endswith(('"', "'")) and message.startswith(('"', "'")):
            message = message[1:-1]

        if path:
            bodydict = dict()
            bodydict["path"] = path
            bodydict["body"] = {"EntryCode": "Maintenance", "Message": message}

            LOGGER.info(
                "Writing maintenance post to %s with %s",
                str(path),
                str(bodydict["body"]),
            )

            self.rdmc.app.post_handler(path, bodydict["body"])

    def repairlogentry(self, options, path=None):
        """Worker function to repair maintenance log

        :param options: command line options
        :type options: list.
        :param path: path of IML log Entries
        :type path: str
        """
        if options.repiml is None or (options.clearlog and options.repiml):
            raise InvalidCommandLineError("")

        if options.service != "IML":
            raise InvalidCommandLineError("Log opted cannot repair maintenance entries.")

        imlid = options.repiml

        if imlid.endswith(('"', "'")) and imlid.startswith(('"', "'")):
            imlid = imlid[1:-1]

        if path:
            imlidpath = [path if path[-1] == "/" else path + "/"][0] + str(imlid) + "/"
            bodydict = dict()
            bodydict["path"] = imlidpath
            if self.rdmc.app.typepath.defs.isgen9:
                bodydict["body"] = {"Oem": {"Hp": {"Repaired": True}}}
            else:
                bodydict["body"] = {"Oem": {"Hpe": {"Repaired": True}}}
            LOGGER.info(
                "Repairing maintenance log at %s with %s",
                str(imlidpath),
                str(bodydict["body"]),
            )

            self.rdmc.app.patch_handler(imlidpath, bodydict["body"])

    def clearlog(self, path):
        """Worker function to clear logs.

        :param path: path to post clear log action
        :type path: str
        """
        LOGGER.info("Clearing logs.")
        if path and self.rdmc.app.typepath.defs.isgen9:
            if path.endswith("/Entries"):
                path = path[: -len("/Entries")]
            elif path.endswith("Entries/"):
                path = path[: -len("Entries/")]

            bodydict = dict()
            bodydict["path"] = path
            bodydict["body"] = {"Action": "ClearLog"}
            self.rdmc.app.post_handler(path, bodydict["body"])
        elif path:
            action = path.split("/")[-2]
            bodydict = dict()
            bodydict["path"] = path
            bodydict["body"] = {"Action": action}
            self.rdmc.app.post_handler(path, bodydict["body"])

    def downloaddata(self, path=None, options=None):
        """Worker function to download the log files

        :param options: command line options
        :type options: list.
        :param path: path to download logs
        :type path: str
        """
        if path:
            LOGGER.info("Getting data from %s", str(path))
            if options.service == "AHS":
                data = self.rdmc.app.get_handler(path, silent=True, uncache=True)
                if data:
                    return data.ori
                else:
                    raise NoContentsFoundForOperationError("Unable to retrieve AHS logs.")
            else:
                data = self.rdmc.app.get_handler(path, silent=True)

            datadict = data.dict

            try:
                completedatadictlist = datadict["Items"] if "Items" in datadict else datadict["Members"]
            except:
                self.rdmc.ui.error("No data available within log.\n")
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")

            if self.rdmc.app.typepath.defs.flagforrest:
                morepages = True

                while morepages:
                    if "links" in datadict and "NextPage" in datadict["links"]:
                        next_link_uri = path + "?page=" + str(datadict["links"]["NextPage"]["page"])

                        href = "%s" % next_link_uri
                        data = self.rdmc.app.get_handler(href, silent=True)
                        datadict = data.dict

                        try:
                            completedatadictlist = completedatadictlist + datadict["Items"]
                        except:
                            self.rdmc.ui.error("No data available within log.\n")
                            raise NoContentsFoundForOperationError("Unable to retrieve logs.")
                    else:
                        morepages = False
            else:
                datadict = list()

                for members in completedatadictlist:
                    if len(list(members.keys())) == 1:
                        memberpath = members[self.rdmc.app.typepath.defs.hrefstring]
                        data = self.rdmc.app.get_handler(memberpath, silent=True)
                        datadict = datadict + [data.dict]
                    else:
                        datadict = datadict + [members]
                completedatadictlist = datadict

            if completedatadictlist:
                try:
                    return completedatadictlist
                except Exception:
                    self.rdmc.ui.error("Could not get the data from server.\n")
                    raise NoContentsFoundForOperationError("Unable to retrieve logs.")
            else:
                self.rdmc.ui.error("No log data present.\n")
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")
        else:
            self.rdmc.ui.error("Path not found for input log.\n")
            raise NoContentsFoundForOperationError("Unable to retrieve logs.")

    def returnimlpath(self, options=None):
        """Return the requested path of the IML logs

        :param options: command line options
        :type options: list.
        """
        LOGGER.info("Obtaining IML path for download.")
        path = ""
        val = "/redfish/v1/Systems/1/LogServices/IML/"
        filtereddatainstance = self.rdmc.app.get_handler(val, silent=True).dict

        try:
            filtereddictslists = filtereddatainstance
            if not filtereddictslists:
                raise NoContentsFoundForOperationError("Unable to retrieve instance.")
            # for filtereddict in filtereddictslists:
            if filtereddictslists["Name"] == "Integrated Management Log":
                if options.clearlog:
                    if self.rdmc.app.typepath.defs.flagforrest:
                        linkpath = filtereddictslists["links"]
                        selfpath = linkpath["self"]
                        path = selfpath["href"]
                    elif self.rdmc.app.typepath.defs.isgen9:
                        path = filtereddictslists[self.rdmc.app.typepath.defs.hrefstring]
                    else:
                        actiondict = filtereddictslists["Actions"]
                        clearkey = [x for x in actiondict if x.endswith("ClearLog")]
                        path = actiondict[clearkey[0]]["target"]
                else:
                    linkpath = filtereddictslists["links"] if "links" in filtereddictslists else filtereddictslists
                    dictpath = linkpath["Entries"]
                    dictpath = dictpath[0] if isinstance(dictpath, list) else dictpath
                    path = dictpath[self.rdmc.app.typepath.defs.hrefstring]
            if not path:
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")
        except NoContentsFoundForOperationError as excp:
            raise excp
        except Exception:
            self.rdmc.ui.error("No path found for the entry.\n")
            raise NoContentsFoundForOperationError("Unable to retrieve logs.")

        if self.rdmc.opts.verbose:
            self.rdmc.ui.printer(str(path) + "\n")

        return path

    def returnielpath(self, options=None):
        """Return the requested path of the IEL logs

        :param options: command line options
        :type options: list.
        """
        LOGGER.info("Obtaining IEL path for download.")
        path = ""
        val = "/redfish/v1/Managers/1/LogServices/IEL/"
        filtereddatainstance = self.rdmc.app.get_handler(val, silent=True).dict

        try:
            filtereddictslists = filtereddatainstance
            if not filtereddictslists:
                raise NoContentsFoundForOperationError("Unable to retrieve instance.")
            # for filtereddict in filtereddictslists:
            if filtereddictslists["Name"] == "iLO Event Log":
                if options.clearlog:
                    if self.rdmc.app.typepath.defs.flagforrest:
                        linkpath = filtereddictslists["links"]
                        selfpath = linkpath["self"]
                        path = selfpath["href"]
                    elif self.rdmc.app.typepath.defs.isgen9:
                        path = filtereddictslists[self.rdmc.app.typepath.defs.hrefstring]
                    else:
                        actiondict = filtereddictslists["Actions"]
                        clearkey = [x for x in actiondict if x.endswith("ClearLog")]
                        path = actiondict[clearkey[0]]["target"]
                else:
                    linkpath = filtereddictslists["links"] if "links" in filtereddictslists else filtereddictslists
                    dictpath = linkpath["Entries"]
                    dictpath = dictpath[0] if isinstance(dictpath, list) else dictpath
                    path = dictpath[self.rdmc.app.typepath.defs.hrefstring]

            if not path:
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")
        except NoContentsFoundForOperationError as excp:
            raise excp
        except Exception:
            self.rdmc.ui.error("No path found for the entry.\n")
            raise NoContentsFoundForOperationError("Unable to retrieve logs.")

        if self.rdmc.opts.verbose:
            self.rdmc.ui.printer(str(path) + "\n")

        return path

    def returnslpath(self, options=None):
        """Return the requested path of the SL logs

        :param options: command line options
        :type options: list.
        """

        if not self.rdmc.app.typepath.defs.isgen10 or not self.rdmc.app.getiloversion() >= 5.210:
            raise IncompatibleiLOVersionError("Security logs are only available on iLO 5 2.10 or" " greater.")
        LOGGER.info("Obtaining SL path for download.")
        path = ""
        val = "/redfish/v1/Systems/1/LogServices/SL/"
        filtereddatainstance = self.rdmc.app.get_handler(val, silent=True).dict

        try:
            filtereddictslists = filtereddatainstance
            if not filtereddictslists:
                raise NoContentsFoundForOperationError("Unable to retrieve instance.")
            # for filtereddict in filtereddictslists:
            if filtereddictslists["Id"] == "SL":
                if options.clearlog:
                    actiondict = filtereddictslists["Actions"]
                    clearkey = [x for x in actiondict if x.endswith("ClearLog")]
                    path = actiondict[clearkey[0]]["target"]
                else:
                    dictpath = filtereddictslists["Entries"]
                    dictpath = dictpath[0] if isinstance(dictpath, list) else dictpath
                    path = dictpath[self.rdmc.app.typepath.defs.hrefstring]

            if not path:
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")
        except NoContentsFoundForOperationError as excp:
            raise excp
        except Exception:
            self.rdmc.ui.error("No path found for the entry.\n")
            raise NoContentsFoundForOperationError("Unable to retrieve logs.")

        if self.rdmc.opts.verbose:
            self.rdmc.ui.printer(str(path) + "\n")

        return path

    def returnahspath(self, options):
        """Return the requested path of the AHS logs

        :param options: command line options
        :type options: list.
        """
        LOGGER.info("Obtaining AHS path for download.")
        path = ""

        if options.filename:
            raise InvalidCommandLineError(
                "AHS logs must be downloaded with " "default name. Please re-run command without filename option."
            )

        val = "/redfish/v1/Managers/1/ActiveHealthSystem/"
        filtereddatainstance = self.rdmc.app.get_handler(val, silent=True).dict

        try:
            filtereddictslists = filtereddatainstance

            if not filtereddictslists:
                raise NoContentsFoundForOperationError("Unable to retrieve log instance.")

            # for filtereddict in filtereddictslists:
            if options.clearlog:
                if self.rdmc.app.typepath.defs.flagforrest:
                    linkpath = filtereddictslists["links"]
                    selfpath = linkpath["self"]
                    path = selfpath["href"]
                elif self.rdmc.app.typepath.defs.isgen9:
                    path = filtereddictslists[self.rdmc.app.typepath.defs.hrefstring]
                else:
                    actiondict = filtereddictslists["Actions"]
                    clearkey = [x for x in actiondict if x.endswith("ClearLog")]
                    path = actiondict[clearkey[0]]["target"]
            else:
                linkpath = filtereddictslists["links"] if "links" in filtereddictslists else filtereddictslists["Links"]

                ahslocpath = linkpath["AHSLocation"]
                path = ahslocpath["extref"]
                weekpath = None

                if options.downloadallahs:
                    path = path
                elif options.customiseAHS:
                    custr = options.customiseAHS
                    if custr.startswith(("'", '"')) and custr.endswith(("'", '"')):
                        custr = custr[1:-1]

                    if custr.startswith("from="):
                        path = path.split("downloadAll=1")[0]

                    path = path + custr
                else:
                    if "RecentWeek" in list(linkpath.keys()):
                        weekpath = linkpath["RecentWeek"]["extref"]
                    elif "AHSFileStart" in list(filtereddictslists.keys()):
                        enddate = filtereddictslists["AHSFileEnd"].split("T")[0]
                        startdate = filtereddictslists["AHSFileStart"].split("T")[0]

                        enddat = list(map(int, enddate.split("-")))
                        startdat = list(map(int, startdate.split("-")))

                        weekago = datetime.datetime.now() - datetime.timedelta(days=6)
                        weekagostr = list(map(int, (str(weekago).split()[0]).split("-")))

                        strdate = min(
                            max(
                                datetime.date(weekagostr[0], weekagostr[1], weekagostr[2]),
                                datetime.date(startdat[0], startdat[1], startdat[2]),
                            ),
                            datetime.date(enddat[0], enddat[1], enddat[2]),
                        )

                        aweekstr = "from=" + str(strdate) + "&&to=" + enddate
                    else:
                        week_ago = datetime.datetime.now() - datetime.timedelta(days=6)
                        aweekstr = (
                            "from=" + str(week_ago).split()[0] + "&&to=" + str(datetime.datetime.now()).split()[0]
                        )

                    path = path.split("downloadAll=1")[0]
                    path = weekpath if weekpath else path + aweekstr

            if not path:
                raise NoContentsFoundForOperationError("Unable to retrieve logs.")
        except NoContentsFoundForOperationError as excp:
            raise excp
        except Exception:
            self.rdmc.ui.error("No path found for the entry.\n")
            raise NoContentsFoundForOperationError("Unable to retrieve logs.")

        if self.rdmc.opts.verbose:
            self.rdmc.ui.printer(str(path) + "\n")

        return path

    def savedata(self, options=None, data=None):
        """Save logs into the specified filename

        :param options: command line options
        :type options: list.
        :param data: log data
        :type data: dict
        """
        LOGGER.info("Saving/Writing data...")
        if data:
            data = self.filterdata(data=data, tofilter=options.filter)
            if options.service == "AHS":
                filename = self.getahsfilename(options)

                with open(filename, "wb") as foutput:
                    foutput.write(data)
            elif options.filename:
                with open(options.filename[0], "w") as foutput:
                    if options.json:
                        foutput.write(str(json.dumps(data, indent=2, sort_keys=True)))
                    else:
                        foutput.write(str(json.dumps(data)))
            else:
                if options.json:
                    UI().print_out_json(data)
                else:
                    UI().print_out_human_readable(data)

    def downloadahslocally(self, options=None):
        """Download AHS logs locally

        :param options: command line options
        :type options: list.
        """
        try:
            self.downloadahslocalworker(options)
        except:
            self.unmountbb()
            raise
        return

    def downloadahslocalworker(self, options):
        """Worker function to download AHS logs locally

        :param options: command line options
        :type options: list.
        """
        LOGGER.info("Entering AHS local download functions...")
        self.dontunmount = True

        if self.rdmc.app.typepath.ilogen and self.rdmc.app.typepath.ilogen < 4:
            raise IncompatibleiLOVersionError("Need at least iLO 4 for this program to run!\n")

        if sys.platform == "darwin":
            raise InvalidCommandLineError("AHS loacal download is not supported on MacOS")
        elif "VMkernel" in platform.uname():
            raise InvalidCommandLineError("AHS loacal download is not supported on VMWare")

        if options.filename:
            raise InvalidCommandLineError(
                "AHS logs must be downloaded with default name! Re-run command without filename!"
            )

        secstate = risblobstore2.BlobStore2().get_security_state()
        # self.rdmc.ui.printer('Security State is {}...\n'.format(secstate))

        if isinstance(secstate, bytes):
            secstate = secstate.decode("utf-8")
        if isinstance(secstate, str):
            secstate = secstate.replace("\x00", "").strip()
        if secstate == "":
            secstate = 0
        self.rdmc.ui.printer("Security State is {}...\n".format(secstate))
        if int(secstate) > 3:
            raise SecurityStateError("AHS logs cannot be downloaded locally in high security state.\n")

        self.lib = risblobstore2.BlobStore2.gethprestchifhandle()

        self.rdmc.ui.printer("Mounting AHS partition...\n")

        try:
            (manual_ovr, abspath) = self.getbbabspath()
        except (PartitionMoutingError, IOError):
            self.mountbb()
            (manual_ovr, abspath) = self.getbbabspath()
            self.dontunmount = False

        LOGGER.info("Blackbox folder path:%s", ",".join(next(os.walk(abspath))[2]))
        if "data" not in abspath:
            self.abspath = os.path.join(abspath, "data")
        LOGGER.info("Blackbox data files path:%s", self.abspath)

        if self.rdmc.app.typepath.ilogen:
            self.updateiloversion()
        try:
            cfilelist = self.getclistfilelisting()
            allfiles = self.getfilenames(options=options, cfilelist=cfilelist)
            self.getdatfilelisting(cfilelist=cfilelist, allfile=allfiles)
            self.createahsfile(ahsfile=self.getahsfilename(options))
        except Exception as excp:
            raise PartitionMoutingError(
                "An exception occurred obtaining Blackbox data files. The directory may be empty: %s.\n" % excp
            )

        if not manual_ovr:
            self.unmountbb()
        else:
            self.unmountbb()
            self.manualunmountbb(abspath)

    def updateiloversion(self):
        """Update iloversion to create appropriate headers."""
        LOGGER.info("Updating iloversion to format data appropriately")
        self.lib.updateiloversion.argtypes = [ctypes.c_float]
        self.lib.updateiloversion(float("2." + str(self.rdmc.app.typepath.ilogen)))

    def createahsfile(self, ahsfile=None):
        """Create the AHS file

        :param ahsfile: ahsfilename
        :type ahsfile: str
        """
        LOGGER.info("Creating AHS file from the formatted data.")
        self.clearahsfile(ahsfile=ahsfile)
        self.lib.setAHSFilepath.argtypes = [ctypes.c_char_p]
        self.lib.setAHSFilepath(os.path.abspath(ahsfile).encode("utf-8", "ignore"))
        self.lib.setBBdatapath.argtypes = [ctypes.c_char_p]
        self.lib.setBBdatapath(self.abspath.encode("utf-8", "ignore"))
        self.lib.createAHSLogFile_G9()

    def clearahsfile(self, ahsfile=None):
        """Clear the ahslog file if already present in filesystem

        :param ahsfile: ahsfilename
        :type ahsfile: str
        """
        LOGGER.info("Clear redundant AHS file in current folder.")
        try:
            os.remove(ahsfile)
        except:
            pass

    def getdatfilelisting(self, cfilelist=None, allfile=None):
        """Create headers based on the AHS log files within blackbox

        :param cfilelist: configuration files in blackbox
        :type cfilelist: list of strings
        :param allfile: all files within blackbox
        :type allfile: list
        """
        LOGGER.info("Reading Blackbox to determine data files.")
        allfile = list(set(allfile) | set(cfilelist))
        LOGGER.info("Final filelist %s", str(allfile))
        for files in allfile:
            if files.startswith((".", "..")):
                continue

            bisrequiredfile = False

            if files.split(".")[0] in [x.split(".")[0] for x in cfilelist]:
                if files.endswith("bb"):
                    bisrequiredfile = True
                    self.lib.updatenfileoptions()

            self.lib.gendatlisting.argtypes = [
                ctypes.c_char_p,
                ctypes.c_bool,
                ctypes.c_uint,
            ]

            filesize = os.stat(os.path.join(self.abspath, files)).st_size
            self.lib.gendatlisting(files.encode("ascii", "ignore"), bisrequiredfile, filesize)

    def getfilenames(self, options=None, cfilelist=None):
        """Get all file names from the blackbox directory."""
        LOGGER.info("Obtaining all relevant file names from Blackbox.")
        datelist = list()
        allfiles = list()

        filenames = next(os.walk(self.abspath))[2]
        timenow = (str(datetime.datetime.now()).split()[0]).split("-")
        strdate = enddate = datetime.date(int(timenow[0]), int(timenow[1]), int(timenow[2]))

        if not options.downloadallahs and not options.customiseAHS:
            weekago = datetime.datetime.now() - datetime.timedelta(days=6)
            weekagostr = (str(weekago).split()[0]).split("-")
            strdate = datetime.date(int(weekagostr[0]), int(weekagostr[1]), int(weekagostr[2]))

        if options.customiseAHS:
            instring = options.customiseAHS
            if instring.startswith(("'", '"')) and instring.endswith(("'", '"')):
                instring = instring[1:-1]
            try:
                (strdatestr, enddatestr) = [e.split("-") for e in instring.split("from=")[-1].split("&&to=")]
                strdate = datetime.date(int(strdatestr[0]), int(strdatestr[1]), int(strdatestr[2]))
                enddate = datetime.date(int(enddatestr[0]), int(enddatestr[1]), int(enddatestr[2]))
            except Exception as excp:
                LOGGER.warning(excp)
                raise InvalidCommandLineError("Cannot parse customised AHSinput.")

        atleastonefile = False
        for files in list(filenames):
            if not files.endswith("bb"):
                # Check the logic for the non bb files
                allfiles.append(files)
                continue

            if options.downloadallahs:
                atleastonefile = True
            if files in ("ilo_boot_support.zbb", "sys_boot_support.zbb"):
                allfiles.append(files)
                filenames.append(files)
                LOGGER.info("%s, number of files%s", files, len(filenames))
                continue
            filenoext = files.rsplit(".", 1)[0]
            filesplit = filenoext.split("-")

            try:
                newdate = datetime.date(int(filesplit[1]), int(filesplit[2]), int(filesplit[3]))
                datelist.append(newdate)

                if (strdate <= newdate) and (newdate <= enddate) and not options.downloadallahs:
                    allfiles.append(files)
                    atleastonefile = True
            except:
                pass

        _ = [cfilelist.remove(fil) for fil in list(cfilelist) if fil not in filenames]

        if options.customiseAHS:
            for files in reversed(cfilelist):
                try:
                    filenoext = files.rsplit(".", 1)[0]
                    filesplit = filenoext.split("-")
                    newdate = datetime.date(int(filesplit[1]), int(filesplit[2]), int(filesplit[3]))
                    if not ((strdate <= newdate) and (newdate <= enddate)):
                        cfilelist.remove(files)
                except:
                    pass

        if options.downloadallahs:
            strdate = min(datelist) if datelist else strdate
            enddate = max(datelist) if datelist else enddate
        else:
            strdate = max(min(datelist), strdate) if datelist else strdate
            enddate = min(max(datelist), enddate) if datelist else enddate

        LOGGER.info("All filenames: %s; Download files: %s", str(filenames), str(allfiles))

        if atleastonefile:
            self.updateminmaxdate(strdate=strdate, enddate=enddate)
            return filenames if options.downloadallahs else allfiles
        else:
            raise NoContentsFoundForOperationError("No AHS log files found.")

    def updateminmaxdate(self, strdate=None, enddate=None):
        """Get the minimum and maximum date of files into header

        :param strdate: starting date of ahs logs
        :type strdate: dateime obj
        :param enddate: ending date of ahs logs
        :type enddate: datetime obj
        """
        LOGGER.info("Updating min and max dates for download.")
        self.lib.updateMinDate.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
        self.lib.updateMinDate(strdate.year, strdate.month, strdate.day)
        self.lib.updateMaxDate.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int]
        self.lib.updateMaxDate(enddate.year, enddate.month, enddate.day)

    def getclistfilelisting(self):
        """Get files present within clist.pkg ."""
        LOGGER.info("Getting all config files that are required.")
        sclistpath = os.path.join(self.abspath, "clist.pkg")
        cfilelist = []
        if os.path.isfile(sclistpath):
            cfile = open(sclistpath, "rb")
            data = cfile.read()

            if data == "":
                raise InvalidCListFileError("Could not read Cfile\n")

            sizeofcfile = len(str(data))
            sizeofrecord = self.lib.sizeofchifbbfilecfgrecord()
            count = sizeofcfile / sizeofrecord
            revcount = 0

            while count >= 1:
                dat = data[revcount * sizeofrecord : (revcount + 1) * sizeofrecord]
                dat = ctypes.create_string_buffer(dat)
                self.lib.getbbfilecfgrecordname.argtypes = [ctypes.c_char_p]
                self.lib.getbbfilecfgrecordname.restype = ctypes.c_char_p
                ptrname = self.lib.getbbfilecfgrecordname(dat)
                name = str(bytearray(ptrname[:32][:]))

                if name not in cfilelist:
                    cfilelist.append(name)

                count = count - 1
                revcount = revcount + 1
        else:
            cfilelist = [f for f in os.listdir(self.abspath) if f.endswith(".zbb")]

        LOGGER.info("CLIST files %s", str(cfilelist))
        return cfilelist

    def getbbabspath(self):
        """Get blackbox folder path."""
        LOGGER.info("Obtaining the absolute path of blackbox.")
        count = 0

        while count < 20:
            if os.name == "nt":
                drives = self.get_available_drives()

                for i in drives:
                    try:
                        label = win32api.GetVolumeInformation(i + ":")[0]

                        if label == "BLACKBOX":
                            abspathbb = i + ":\\data\\"
                            self.abspath = abspathbb
                            cfilelist = self.getclistfilelisting()
                            if not cfilelist:
                                self.unmountbb()
                                self.manualmountbb()
                            else:
                                return False, abspathbb
                    except:
                        pass
            else:
                with open("/proc/mounts", "r") as fmount:
                    while True:
                        lin = fmount.readline()

                        if len(lin.strip()) == 0:
                            break

                        if r"/BLACKBOX" in lin:
                            abspathbb = lin.split()[1]
                            self.abspath = abspathbb
                            cfilelist = self.getclistfilelisting()
                            if not cfilelist:
                                self.unmountbb()
                                self.manualmountbb()
                            else:
                                return False, abspathbb

                if count > 5:
                    found, path = self.manualmountbb()
                    if found:
                        return True, path

            count = count + 1
            time.sleep(5)

        raise PartitionMoutingError("iLO not responding to request " "for mounting AHS partition")

    def manualmountbb(self):
        """Manually mount blackbox when after fixed time."""
        LOGGER.info("Manually mounting the blackbox.")

        try:
            context = pyudev.Context()
            for device in context.list_devices(subsystem="block"):
                if device.get("ID_FS_LABEL") == "BLACKBOX":
                    dirpath = os.path.join(tempfile.gettempdir(), "BLACKBOX")

                    if not os.path.exists(dirpath):
                        try:
                            os.makedirs(dirpath)
                        except Exception as excp:
                            raise excp

                    pmount = subprocess.Popen(
                        ["mount", device.device_node, dirpath],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                    )
                    _, _ = pmount.communicate()

                    return True, dirpath
        except UnicodeDecodeError as excp:
            self.unmountbb()
            raise UnabletoFindDriveError(excp)

        return False, None

    def manualunmountbb(self, dirpath):
        """Manually unmount blackbox when after fixed time

        :param dirpath: mounted directory path
        :type dirpath: str
        """
        LOGGER.info("Manually unmounting the blackbox.")
        pmount = subprocess.Popen(["umount", dirpath], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        _, _ = pmount.communicate()

    def mountbb(self):
        """Mount blackbox."""
        LOGGER.info("Mounting blackbox...")
        bs2 = risblobstore2.BlobStore2()
        bs2.mount_blackbox()
        bs2.channel.close()

    def unmountbb(self):
        """Unmount blacbox."""
        if not self.dontunmount:
            bs2 = risblobstore2.BlobStore2()
            bs2.bb_media_unmount()
            bs2.channel.close()

    def get_available_drives(self):
        """Obtain all drives"""
        LOGGER.info("Unmounting blackbox...")
        if "Windows" not in platform.system():
            return []

        drive_bitmask = ctypes.cdll.kernel32.GetLogicalDrives()

        return list(
            itertools.compress(
                string.ascii_uppercase,
                [ord(drive) - ord("0") for drive in bin(drive_bitmask)[:1:-1]],
            )
        )

    def filterdata(self, data=None, tofilter=None):
        """Filter the logs

        :param data: log data
        :type data: dict
        :param tofilter: command line filter option
        :type tofilter: str
        """
        LOGGER.info("Filtering logs based on requsted options.")
        if tofilter and data:
            try:
                if (str(tofilter)[0] == str(tofilter)[-1]) and str(tofilter).startswith(("'", '"')):
                    tofilter = tofilter[1:-1]

                (sel, val) = tofilter.split("=")
                sel = sel.strip()
                val = val.strip()

                if val.lower() == "true" or val.lower() == "false":
                    val = val.lower() in ("yes", "true", "t", "1")
            except:
                raise InvalidCommandLineError("Invalid filter" " parameter format [filter_attribute]=[filter_value]")

            # Severity inside Oem/Hpe should be filtered
            if sel == "Severity":
                if not self.rdmc.app.typepath.defs.isgen9:
                    sel = "Oem/Hpe/" + sel
                else:
                    sel = "/" + sel
            data = filter_output(data, sel, val)
            if not data:
                raise NoContentsFoundForOperationError("Filter returned no matches.")

        return data

    def getahsfilename(self, options):
        """Create a default name if no ahsfilename is passed

        :param options: command line options
        :type options: list.
        """
        LOGGER.info("Obtaining Serialnumber from iLO for AHS filename.")

        if not options.sessionid:
            val = "ComputerSystem."
            filtereddatainstance = self.rdmc.app.select(selector=val)
            snum = None

            try:
                filtereddictslists = [x.resp.dict for x in filtereddatainstance]

                if not filtereddictslists:
                    raise NoContentsFoundForOperationError("")
            except Exception:
                try:
                    resp = self.rdmc.app.get_handler(
                        self.rdmc.app.typepath.defs.systempath,
                        silent=True,
                        service=True,
                        uncache=True,
                    )
                    snum = resp.dict["SerialNumber"] if resp else snum
                except KeyError:
                    raise InvalidKeyError(
                        "Unable to find key SerialNumber, please check path %s" % self.rdmc.app.typepath.defs.systempath
                    )
                except:
                    raise NoContentsFoundForOperationError("Unable to retrieve log instance.")

            snum = filtereddictslists[0]["SerialNumber"] if not snum else snum
        else:
            resp = self.rdmc.app.get_handler(
                "/redfish/v1/systems/1",
                sessionid=options.sessionid,
                silent=True,
                service=True,
                uncache=True,
            )
            if "SerialNumber" in resp.dict:
                snum = resp.dict["SerialNumber"] if resp else snum
        snum = "UNKNOWN" if snum.isspace() else snum
        timenow = (str(datetime.datetime.now()).split()[0]).split("-")
        todaysdate = "".join(timenow)

        resp = self.rdmc.app.get_handler(
            "/redfish/v1",
            sessionid=options.sessionid,
            silent=True,
            service=True,
            uncache=True,
        )
        ahsdefaultfilename = "HPE_" + snum + "_" + todaysdate + ".ahs"
        if "Vendor" in resp.dict:
            ahsdefaultfilename = ahsdefaultfilename.replace("HPE", resp.dict["Vendor"])

        if options.directorypath:
            dir_name = options.directorypath
            dir_name = dir_name.encode("utf-8").decode("utf-8")
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)
            ahsdefaultfilename = os.path.join(dir_name, ahsdefaultfilename)

        return ahsdefaultfilename

    def serverlogsvalidation(self, options):
        """Serverlogs method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-f",
            "--filename",
            dest="filename",
            help="Use this flag if you wish to use a different"
            " filename than the default one. The default filename is"
            " ilorest.json.",
            action="append",
            default=None,
        )
        customparser.add_argument(
            "--filter",
            dest="filter",
            help="Optionally set a filter value for a filter attribute."
            " This uses the provided filter for the currently selected"
            " type. Note: Use this flag to narrow down your results. For"
            " example, selecting a common type might return multiple"
            " objects that are all of that type. If you want to modify"
            " the properties of only one of those objects, use the filter"
            " flag to narrow down results based on properties."
            "\t\t\t\t\t Usage: --filter [ATTRIBUTE]=[VALUE]",
            default=None,
        )
        customparser.add_argument(
            "--selectlog",
            dest="service",
            help="""Read log from the given log service. Options: IML, """ """IEL or AHS.""",
            default=None,
        )
        customparser.add_argument(
            "--clearlog",
            "-c",
            dest="clearlog",
            action="store_true",
            help="""Clears the logs for a the selected option.""",
            default=None,
        )
        customparser.add_argument(
            "--maintenancemessage",
            "-m",
            dest="mainmes",
            help="""Maintenance message to be inserted into the log. """ """(IML LOGS ONLY FEATURE)""",
            default=None,
        )
        customparser.add_argument(
            "--customiseAHS",
            dest="customiseAHS",
            help="""Allows customised AHS log data to be downloaded.""",
            default=None,
        )
        customparser.add_argument(
            "--downloadallahs",
            dest="downloadallahs",
            action="store_true",
            help="""Allows complete AHS log data to be downloaded.""",
            default=None,
        )
        customparser.add_argument(
            "--directorypath",
            dest="directorypath",
            help="""Directory path for the ahs file.""",
            default=None,
        )
        customparser.add_argument(
            "--mpfile",
            dest="mpfilename",
            help="""use the provided filename to obtain server information.""",
            default=None,
        )
        customparser.add_argument(
            "--outputdirectory",
            dest="outdirectory",
            help="""use the provided directory to output data for multiple server downloads.""",
            default=None,
        )
        customparser.add_argument(
            "--mplog",
            dest="mplog",
            help="""used to indicate the logs to be downloaded on multiple servers. """
            """Allowable values: IEL, IML, AHS, all or combination of any two.""",
            default=None,
        )
        customparser.add_argument(
            "--repair",
            "-r",
            dest="repiml",
            help="""Repair the IML logs with the given ID.""",
            default=None,
        )
        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Reboot Command for rdmc """

import time

from six.moves import input

try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )


class RebootCommand:
    """Reboot server that is currently logged in"""

    def __init__(self):
        self.ident = {
            "name": "reboot",
            "usage": None,
            "description": "Remotely control system power state commands such as, "
            "\n\t1. Turning the system on.\n\t2. Turning the system off.\n\t3. Power "
            "cycling/rebooting.\n\t4. Issuing a Non-Maskable Interrupt (NMI).\n\t5. Any "
            "number of pre-defined operations through virtual power-button presses."
            "\n\n\tNote: By default a force "
            "restart will occur, if the system is in an applicable power state.\n\texample: "
            "reboot On\n\n\tOPTIONAL PARAMETERS AND DESCRIPTIONS:"
            "\n\tOn \t\t(Turns the system on.)\n\tForceOff  "
            "\t(Performs an immediate non-graceful shutdown.)"
            "\n\tForceRestart \t(DEFAULT) (Performs"
            " an immediate non-graceful shutdown,\n\t\t\t"
            " followed by a restart of the system.)\n\tNmi  "
            "\t\t(Generates a Non-Maskable Interrupt to cause"
            " an\n\t\t\t immediate system halt.)\n\tPushPowerButton "
            "(Simulates the pressing of the physical power "
            "button\n\t\t\t on this system.)\n\n\tOEM PARAMETERS AND"
            " DESCRIPTIONS:\n\tPress\t\t(Simulates the pressing of the"
            " physical power button\n\t\t\t on this system.)\n\t"
            "PressAndHold\t(Simulates pressing and holding of the power"
            " button\n\t\t\t on this systems.)\n\tColdBoot\t(Immidiately"
            " Removes power from the server,\n\t\t\tfollowed by a restart"
            " of the system)",
            "summary": "Reboot operations for the current logged in server.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main reboot worker function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if len(args) < 2:
            self.rebootvalidation(options)
        else:
            raise InvalidCommandLineError("Invalid number of parameters." " Reboot takes a maximum of 1 parameter.")

        if not args:
            self.rdmc.ui.warn(
                "\nAfter the server is rebooted the session will be terminated."
                "\nPlease wait for the server to boot completely before logging in "
                "again.\nRebooting server in 3 seconds...\n"
            )
            time.sleep(3)
        else:
            self.printreboothelp(args[0])
            time.sleep(3)

        select = "ComputerSystem."
        results = self.rdmc.app.select(selector=select)
        oemlist = ["press", "pressandhold", "coldboot"]

        try:
            results = results[0]
        except:
            pass

        if results:
            put_path = results.resp.request.path
        else:
            raise NoContentsFoundForOperationError("Unable to find %s" % select)

        if args and args[0].lower() in oemlist:
            bodydict = results.resp.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]

            if args[0].lower() == "coldboot":
                resettype = "SystemReset"
            else:
                resettype = "PowerButton"
        else:
            bodydict = results.resp.dict
            resettype = "Reset"

        try:
            for item in bodydict["Actions"]:
                if resettype in item:
                    if self.rdmc.app.typepath.defs.isgen10:
                        action = item.split("#")[-1]
                    else:
                        action = resettype

                    put_path = bodydict["Actions"][item]["target"]
                    break
        except:
            action = resettype

        if args and not args[0].lower() == "forcerestart":
            if args[0].lower() == "on":
                body = {"Action": action, "ResetType": "On"}
            elif args[0].lower() == "forceoff":
                body = {"Action": action, "ResetType": "ForceOff"}
            elif args[0].lower() == "nmi":
                body = {"Action": action, "ResetType": "Nmi"}
            elif args[0].lower() == "pushpowerbutton":
                body = {"Action": action, "ResetType": "PushPowerButton"}
            elif args[0].lower() == "press":
                body = {"Action": action, "PushType": "Press"}
            elif args[0].lower() == "pressandhold":
                body = {"Action": action, "PushType": "PressAndHold"}
            elif args[0].lower() == "coldboot":
                body = {"Action": action, "ResetType": "ColdBoot"}
        else:
            body = {"Action": action, "ResetType": "ForceRestart"}

        if options.confirm is True:
            count = 0
            while True:
                count = count + 1
                confirmation = input("Rebooting system, type yes to confirm or no to abort:")
                if confirmation.lower() in ("no", "n") or count > 3:
                    self.rdmc.ui.printer("Aborting reboot.\n")
                    return ReturnCodes.SUCCESS
                elif confirmation.lower() in ("yes", "y"):
                    break

        self.rdmc.app.post_handler(put_path, body)

        if not options.nologout:
            self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def printreboothelp(self, flag):
        """helper print function for reboot function

        :param flag: command line option
        :type flag: str
        """
        if flag.upper() == "ON":
            self.rdmc.ui.warn(
                "\nThe server is powering on. Note, the current session will be "
                "terminated.\nPlease wait for the server to boot completely before logging in "
                "again.\nTurning on the server in 3 seconds...\n"
            )
        elif flag.upper() == "FORCEOFF":
            self.rdmc.ui.warn(
                "\nThe server is powering off. Note, the current session will be "
                "terminated.\nPlease wait for the server to power off completely before logging "
                "in again.\nPowering off the server in 3 seconds...\n"
            )
        elif flag.upper() == "FORCERESTART":
            self.rdmc.ui.warn(
                "\nForcing a server restart. Note, the current session will be "
                "terminated.\nPlease wait for the server to boot completely before logging in "
                "again.\nRebooting the server in 3 seconds...\n"
            )
        elif flag.upper() == "NMI":
            self.rdmc.ui.warn(
                "\nA non-maskable interrupt will be issued to this server. Note, the "
                "current session will be terminated.\nIssuing interrupt in 3 seconds...\n"
            )
        elif flag.upper() == "PUSHPOWERBUTTON" or flag.upper() == "PRESS":
            self.rdmc.ui.warn(
                "\nThe server power button will be virtually pushed; the reaction "
                "will be dependent on the current system power and boot state. Note the current "
                "session will be terminated.\nVirtual push in 3 seconds...\n"
            )
        elif flag.upper() == "COLDBOOT":
            self.rdmc.ui.warn(
                "\nThe server will be cold boot, power cycled. Note, the current "
                "session will be terminated.\nPlease wait for the server to boot completely "
                "before logging in again.\nCold Booting server in 3 seconds...\n"
            )
        elif flag.upper() == "PRESSANDHOLD":
            self.rdmc.ui.warn(
                "\nThe server will be forcefully powered off. Note, the current "
                "session will be terminated.\nPlease wait for the server to power off "
                "completely before logging in again.\nPressing and holding the power button in "
                "3 seconds...\n"
            )
        else:
            raise InvalidCommandLineError("Invalid parameter: '%s'. Please run" " 'help reboot' for parameters." % flag)

    def rebootvalidation(self, options):
        """reboot method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--confirm",
            dest="confirm",
            action="store_true",
            help="Optionally include to request user confirmation for reboot.",
            default=False,
        )
        customparser.add_argument(
            "--nologout",
            dest="nologout",
            action="store_true",
            help="Optionally include to not to logout of iLO connection.",
            default=False,
        )



###
# Copyright 2016-2022 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Add License Command for rdmc """

try:
    from rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        PathUnavailableError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineErrorOPTS,
        PathUnavailableError,
        IncompatibleiLOVersionError,
    )

from redfish.ris import rmc_helper


class IloLicenseCommand:
    """Add an iLO license to the server"""

    def __init__(self):
        self.ident = {
            "name": "ilolicense",
            "usage": None,
            "description": "Set an iLO license on the current logged in server.\n\t",
            "summary": "Adds an iLO license key to the currently logged in server.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main ilolicense Function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("Provide required argument")

        self.addlicensevalidation(options)

        code = self.ilolicenseworkerfunction(options, args)

        self.cmdbase.logout_routine(self, options)
        # Return code
        return code

    def ilolicenseworkerfunction(self, options, args):
        """
        Ilolicense worker function. It calls appropriate function.
        :param options: command line options
        :type options: list.
        :param args: command line args
        :type args: string.
        """
        path = self.rdmc.app.typepath.defs.addlicensepath
        if options.check_license:
            self.check_license(options, path)
            return ReturnCodes.SUCCESS
        if options.check_state:
            if self.rdmc.app.typepath.defs.isgen10:
                result = self.get_license(options, path)
                state = result.dict["ConfirmationRequest"]["EON"]["State"]
                self.rdmc.ui.print_out_json("State: " + state)
                if state == "confirmed":
                    self.rdmc.ui.printer("License is confirmed\n")
                if state == "unconfirmed":
                    self.rdmc.ui.printer("License is not confirmed\n")
                if state == "unlicensed":
                    self.rdmc.ui.printer("Server is unlicensed\n")
                if state == "evaluation":
                    self.rdmc.ui.printer("Server is in evaluation mode\n")
                return ReturnCodes.SUCCESS
            else:
                self.rdmc.ui.printer("Feature supported only for Gen 10 and above\n")
                return ReturnCodes.SUCCESS
        if options.license_key:
            return_code = self.license_key(options, path, args)
            return return_code
        if options.uninstall_license:
            return_code = self.uninstall_license(options, path)
            return return_code
        if (
            len(args) == 0
            and options.license_key is None
            and options.check_license is None
            and options.check_state is False
            and options.uninstall_license is False
        ):
            result = self.get_license(options, path)
            self.print_license_info(result.dict)
            return ReturnCodes.SUCCESS
        if (
            len(args) == 1
            and options.license_key is None
            and options.check_license is None
            and options.check_state is False
            and options.uninstall_license is False
        ):
            return_code = self.license_key(options, path, args)
            return return_code

    def print_license_info(self, results):
        """
        Prints the license info
        """
        for key, value in results.items():
            if "@odata" not in key:
                if type(value) is dict:
                    self.print_license_info(value)
                else:
                    self.rdmc.ui.printer(key + ":" + str(value) + "\n")

    def uninstall_license(self, options, path):
        """
        Deletes the license
        """
        if self.rdmc.app.typepath.defs.isgen10:
            path = path + "1/"
        else:
            path = path + "/1"
        try:
            results = self.rdmc.app.delete_handler(path, silent=True)
            if results.status == 404:
                raise PathUnavailableError("License is not installed")
                # return ReturnCodes.SUCCESS
            if results.status == 403:
                self.rdmc.ui.error("Insufficient Privilege to uninstall license")
                return ReturnCodes.RIS_MISSING_ID_TOKEN
            if results.status == 400:
                self.rdmc.ui.printer("There is no license available to uninstall\n")
                return ReturnCodes.SUCCESS
            if results.status == 200:
                self.rdmc.ui.printer("Uninstalled license successfully\n")
                return ReturnCodes.SUCCESS
        except rmc_helper.IloLicenseError:
            self.rdmc.ui.error("Error Occured while Uninstall")
            return ReturnCodes.ILO_LICENSE_ERROR
        except IncompatibleiLOVersionError:
            self.rdmc.ui.error("iLO FW version on this server doesnt support this operation")
            return ReturnCodes.INCOMPATIBLE_ILO_VERSION_ERROR

    def license_key(self, options, path, args):
        """
        Installs the license
        """
        if options.license_key is not None:
            if options.license_key[0] is not None:
                body = {"LicenseKey": "%s" % options.license_key[0]}
            else:
                self.rdmc.ui.printer("Provide license key to install\n")
        if len(args) == 1 and options.license_key is None:
            body = {"LicenseKey": "%s" % args[0]}
        try:
            results = self.rdmc.app.post_handler(path, body)
            if results.status == 201:
                return ReturnCodes.SUCCESS
        except rmc_helper.IloLicenseError:
            self.rdmc.ui.error("Error Occured while install")
            return ReturnCodes.ILO_LICENSE_ERROR
        except rmc_helper.IdTokenError:
            self.rdmc.ui.error("Insufficient Privilege to update license")
            return ReturnCodes.RIS_MISSING_ID_TOKEN

    def check_license(self, options, path):
        """
        Checks and Displays the license
        """
        if self.rdmc.app.typepath.defs.isgen10:
            if options.check_license[0] is not None:
                result = self.get_license(options, path)

                if result.dict["ConfirmationRequest"]["EON"]["LicenseKey"] == options.check_license[0]:
                    self.rdmc.ui.printer("Matched. Provided key is installed on this server\n")
                else:
                    self.rdmc.ui.printer("Provided key is not installed on this server\n")
            else:
                self.rdmc.ui.printer("Provide license key to check\n")
        else:
            self.rdmc.ui.printer("Feature supported only on gen 10 and above\n")

    def get_license(self, options, path):
        """
        Gets the license
        """
        if self.rdmc.app.typepath.defs.isgen10:
            path = path + "1/"
        else:
            path = path + "/1"
        results = self.rdmc.app.get_handler(path, silent=True)
        if results.status == 200:
            return results

    def addlicensevalidation(self, options):
        """ilolicense validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--install",
            dest="license_key",
            action="append",
            help="""Installs the given normal or premium license""",
            default=None,
        )
        customparser.add_argument(
            "--uninstall",
            help="""Deletes the installed license""",
            action="store_true",
            dest="uninstall_license",
        )
        customparser.add_argument(
            "--check",
            dest="check_license",
            help="""Lists the specified license""",
            action="append",
            default=None,
        )
        customparser.add_argument(
            "--check_confirm",
            dest="check_state",
            action="store_true",
            help="Checks the confirmed state and displays it",
        )



"""Module containing functional tests for the device flag"""
from __future__ import absolute_import

import sys

from rdmc_helper import ReturnCodes


class TestDeviceFlag(object):
    """Functional Tests for the --device | -D flag"""
    @staticmethod
    def test_with_no_dimm_id(ilorest, dl360_large):
        """Test for --device flag with no DIMM IDs specified"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_valid_non_pmem_dimm_id(ilorest, dl360_large):
        """
        Test for --device flag with valid DIMM IDs
        which are not Persistent Memory Modules specified
        """
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device', '--dimm', '1@1,2@3'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_invalid_dimm_id_1(ilorest):
        """
        Test for --device flag with invalid DIMM IDs specified.
        Fails at Validation.
        """
        results = ilorest.run(['showdcpmm', '--device', '--dimm', '132,243,xyz'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_invalid_dimm_id_2(ilorest, dl360_large):
        """
        Test for --device flag with invalid DIMM IDs specified
        Passes Validation, Fails later when DIMM IDs not found
        """
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device', '--dimm', '8@2,2@18'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_valid_pmem_dimm_id(ilorest, dl360_large):
        """
        Test for --device flag with valid DIMM IDs
        which are Persistent Memory Modules specified
        """
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device', '--dimm', '1@2'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_valid_pmem_dimm_id_json(ilorest, dl360_large):
        """
        Test for --device flag with valid DIMM IDs
        which are Persistent Memory Modules specified
        along with the --json flag
        """
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device', '--dimm', '1@2', '--json'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_json(ilorest, dl360_large):
        """
        Test for --device flag with valid DIMM IDs
        which are Persistent Memory Modules specified
        along with the --json flag
        """
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['showdcpmm', '--device', '--json'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.SUCCESS



###
# Copyright 2016 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###


# -*- coding: utf-8 -*-
""" New Command for RDMC """

import json
import struct

import redfish.ris

try:
    from rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )

__filename__ = "smbios.json"


class SMBiosCommand:
    """Main smbios command class"""

    def __init__(self):
        self.ident = {
            "name": "smbios",
            "usage": None,
            "description": "Run to get the smbios for the system." "\n\texample write smbios data to file: smbios",
            "summary": "Gets the smbios for the currently logged in server and "
            "write results to a file in json format.",
            "aliases": [],
            "auxcommands": [],
        }
        self.rdmc = None
        self.filename = None

    def smbiosfunction(self, options):
        """Main smbios command worker function

        :param options: command options
        :type options: options.
        """

        sysresp = self.rdmc.app.get_handler(self.rdmc.app.typepath.defs.systempath, silent=True, service=True)

        try:
            path = sysresp.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]["SMBIOS"]["extref"]
        except:
            raise NoContentsFoundForOperationError("Unable to find ComputerSystems")

        resp = self.rdmc.app.get_handler(path, silent=True, service=True)

        if resp and resp.status == 200:
            data = self.unpackdata(resp.ori)
            data = {"smbios data": data}

            outfile = open(self.filename, "w")
            outfile.write(json.dumps(data, indent=2, cls=redfish.ris.JSONEncoder, sort_keys=True))
            outfile.close()

            self.rdmc.ui.printer("Smbios saved to: %s\n" % self.filename)
        else:
            raise NoContentsFoundForOperationError("Unable to find smbios.")

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def unpackdata(self, data):
        """Unpacks and returns json formatted data

        :param data: data to unpack
        :type data: string.
        """

        dataptr = 0
        fulldata = []
        if not isinstance(data, (bytes, bytearray)):
            data = data.encode()

        try:
            while True:
                record_header = struct.unpack_from("BBH", data[dataptr : dataptr + 4])

                record_type = record_header[0]
                record_length = record_header[1]
                record_handle = record_header[2]

                # skip to the strings
                dataptr += record_length

                # read strings from end of record
                record_strings = []

                while True:
                    tempstr = ""

                    while data[dataptr] != 0:
                        tempstr += chr(data[dataptr])
                        dataptr += 1

                    dataptr += 1

                    if len(tempstr) == 0:
                        break

                    record_strings.append(str(tempstr))

                while data[dataptr] == 0:
                    dataptr += 1

                fulldata.append(
                    {
                        "Type": str(record_type),
                        "Length": str(record_length),
                        "Handle": str(record_handle),
                    }
                )

                if len(record_strings):
                    fulldata[-1]["Strings"] = str(record_strings)
        except IndexError:
            pass
        return fulldata

    def run(self, line, help_disp=False):
        """Wrapper function for smbios command main function

        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if self.rdmc.app.typepath.url:
            if "http" not in self.rdmc.app.typepath.url:
                options.logout = True
                self.cmdbase.logout_routine(self, options)
        else:
            options.logout = True
            self.cmdbase.logout_routine(self, options)
        self.smbiosvalidation(options)

        if not self.rdmc.app.typepath.flagiften:
            raise IncompatibleiLOVersionError("smbios command is RedFish only.")

        self.smbiosfunction(options)
        # Return code
        return ReturnCodes.SUCCESS

    def smbiosvalidation(self, options):
        """smbios command method validation function

        :param options: command options
        :type options: options.
        """
        self.cmdbase.login_select_validation(self, options)

        # filename validations and checks
        self.filename = None

        if options.filename:
            self.filename = options.filename[0]

        if not self.filename:
            self.filename = __filename__

    def definearguments(self, customparser):
        """Wrapper function for smbios command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-f",
            "--filename",
            dest="filename",
            help="Use this flag if you wish to use a different"
            " filename than the default one. The default filename is"
            " %s." % __filename__,
            action="append",
            default=None,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Info Command for RDMC """

try:
    from rdmc_helper import (
        InfoMissingEntriesError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InfoMissingEntriesError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )

try:
    from rdmc_helper import HARDCODEDLIST
except:
    from ilorest.rdmc_helper import HARDCODEDLIST


class InfoCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "info",
            "usage": None,
            "description": "Displays detailed "
            "information about a property within a selected type"
            "\n\texample: info property\n\n\tDisplays detailed "
            "information for several properties\n\twithin a selected "
            "type\n\texample: info property property/sub-property property\n\n\t"
            "Run without arguments to display properties \n\tthat "
            "are available for info command\n\texample: info",
            "summary": "Displays detailed information about a property within a selected type.",
            "aliases": [],
            "auxcommands": [],
        }

        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, autotest=False, help_disp=False):
        """Main info worker function

        :param line: command line input
        :type line: string.
        :param autotest: flag to determine if running automatictesting
        :type autotest: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.infovalidation(options)

        if args:
            for item in args:
                if self.rdmc.app.selector.lower().startswith("bios.") and "attributes" not in item.lower():
                    if not (item.lower() in HARDCODEDLIST or "@odata" in item.lower()):
                        item = "Attributes/" + item

                outdata = self.rdmc.app.info(props=item, dumpjson=options.json, latestschema=options.latestschema)

                if autotest:
                    return outdata
                if outdata and options.json:
                    self.rdmc.ui.print_out_json(outdata)
                elif outdata:
                    self.rdmc.ui.printer(outdata)

                if not outdata:
                    raise InfoMissingEntriesError("There are no valid " "entries for info in the current instance.")
                else:
                    if len(args) > 1 and not item == args[-1]:
                        self.rdmc.ui.printer("\n************************************" "**************\n")
        else:
            results = set()
            instances = self.rdmc.app.select()
            for instance in instances:
                currdict = instance.resp.dict
                currdict = (
                    currdict["Attributes"]
                    if instance.maj_type.startswith(self.rdmc.app.typepath.defs.biostype)
                    and currdict.get("Attributes", None)
                    else currdict
                )
                results.update([key for key in currdict if key not in HARDCODEDLIST and "@odata" not in key.lower()])

            if results and autotest:
                return results
            elif results:
                self.rdmc.ui.printer("Info options:\n")
                for item in results:
                    self.rdmc.ui.printer("%s\n" % item)
            else:
                raise InfoMissingEntriesError(
                    "No info items available for this selected type." " Try running with the --latestschema flag."
                )

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def infovalidation(self, options):
        """Info method validation function

        :param options: command line options
        :type options: list.
        """

        if self.rdmc.opts.latestschema:
            options.latestschema = True

        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--selector",
            dest="selector",
            help="Optionally include this flag to select a type to run"
            " the current command on. Use this flag when you wish to"
            " select a type without entering another command, or if you"
            " wish to work with a type that is different from the one"
            " you currently have selected.",
            default=None,
        )

        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        customparser.add_argument(
            "--latestschema",
            dest="latestschema",
            action="store_true",
            help="Optionally use the latest schema instead of the one "
            "requested by the file. Note: May cause errors in some data "
            "retrieval due to difference in schema versions.",
            default=None,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-

"""  Command to Display Security State of Dimms """

# ---------Imports---------
from __future__ import absolute_import

from tabulate import tabulate

try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
    )

from .lib.RestHelpers import RestHelpers

# ---------End of imports---------


class DisplaySecurityStateCommand:
    """Command to Display Security State of Dimms"""

    def __init__(self):
        self.ident = {
            "name": "pmmsecuritystate",
            "usage": None,
            "description": "Displaying the Security state of dimms\n\texample: pmmsecuritystate",
            "summary": "Displaying the Security state of dimms.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()
        self._rest_helpers = None

    def display_SecurityState(self):
        resp = self._rest_helpers.retrieve_pmem_location()

        pmemLocation = []
        for item in resp["Members"]:
            if item["MemoryType"] == "IntelOptane":
                pmemLocation.append(item["Name"])

        securityState = []
        headers = ["Location", "State"]
        for i in pmemLocation:
            path = "/redfish/v1/Systems/1/Memory/" + i
            resp = self._rest_helpers.retrieve_security_state(path)
            state = resp["SecurityState"]
            securityState.append(state)

        merged_list = tuple(zip(pmemLocation, securityState))
        print(tabulate(merged_list, headers, tablefmt="psql"))

    def run(self, line, help_disp=False):
        """
        Wrapper function for new command main function
        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineError("Failed to parse options")
        if args:
            raise InvalidCommandLineError("Chosen flag doesn't expect additional arguments")

        # Raise exception if server is in POST
        self._rest_helpers = RestHelpers(rdmcObject=self.rdmc)
        if self._rest_helpers.in_post():
            raise NoContentsFoundForOperationError(
                "Unable to retrieve resources - " "server might be in POST or powered off"
            )

        self.display_SecurityState()

        return ReturnCodes.SUCCESS

    def definearguments(self, customparser):
        """Wrapper function for smbios command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)



# ##
# Copyright 2016-2023 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ##

# -*- coding: utf-8 -*-
""" Update Task Queue Command for rdmc """

from argparse import RawDescriptionHelpFormatter
from random import randint

from redfish.ris.rmc_helper import IdTokenError

try:
    from rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
        TaskQueueError,
        LOGGER,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
        TaskQueueError,
        LOGGER,
    )

__subparsers__ = ["create"]


class UpdateTaskQueueCommand:
    """Main download command class"""

    def __init__(self):
        self.ident = {
            "name": "taskqueue",
            "usage": None,
            "description": "Run to add or remove tasks from the task queue. Added tasks are "
            "appended to the end of the queue. Note: iLO 5 required.\n"
            "Example:\n\ttaskqueue create 30\n\t"
            "taskqueue create <COMP_NAME>\n\t"
            "taskqueue --cleanqueue\n\t"
            "taskqueue --resetqueue\n\t"
            "taskqueue\n",
            "summary": "Manages the update task queue for iLO.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main update task queue worker function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            ident_subparser = False
            for cmnd in __subparsers__:
                if cmnd in line:
                    (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
                    ident_subparser = True
                    break
            if not ident_subparser:
                (options, args) = self.rdmc.rdmc_parse_arglist(self, line, default=True)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.updatetaskqueuevalidation(options)

        if self.rdmc.app.typepath.defs.isgen9:
            raise IncompatibleiLOVersionError("iLO Repository commands are only available on iLO 5.")

        if options.command.lower() == "create":
            self.createtask(options.keywords, options)
        else:
            if options.resetqueue:
                self.resetqueue(options)
            elif options.cleanqueue:
                self.cleanqueue()
            self.printqueue(options)

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def resetqueue(self, options):
        """Deletes everything in the update task queue"""
        tasks = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/UpdateTaskQueue/")
        if not tasks:
            self.rdmc.ui.printer("No tasks found.\n")

        self.rdmc.ui.printer("Deleting all update tasks...\n")

        for task in tasks:
            self.rdmc.ui.printer("Deleting: %s\n" % task["Name"])
            if "RecoveryPrivilege" in task and task["RecoveryPrivilege"]:
                self.rdmc.ui.printer("This task is associated with updating recovery set\n")
                if not (options.user and options.password):
                    raise TaskQueueError("Deleting this task needs " "--user and --password options, delete failed\n")
                if "blobstore" in self.rdmc.app.current_client.base_url:
                    LOGGER.info("Logging out of the session without user and password")
                    self.rdmc.app.current_client.logout()
                    LOGGER.info("Logging in with user and password for deleting system recovery set")
                    self.rdmc.app.current_client._user_pass = (options.user, options.password)
                    self.rdmc.app.current_client.login(self.rdmc.app.current_client.auth_type)
                    self.rdmc.app.delete_handler(task["@odata.id"])
                else:
                    self.rdmc.app.delete_handler(task["@odata.id"])
            else:
                self.rdmc.app.delete_handler(task["@odata.id"])

    def cleanqueue(self):
        """Deletes all finished or errored tasks in the update task queue"""
        tasks = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/UpdateTaskQueue/")
        if not tasks:
            self.rdmc.ui.printer("No tasks found.\n")

        self.rdmc.ui.printer("Cleaning update task queue...\n")

        for task in tasks:
            if task["State"] == "Complete" or task["State"] == "Exception":
                self.rdmc.ui.printer("Deleting %s...\n" % task["Name"])
                self.rdmc.app.delete_handler(task["@odata.id"])

    def createtask(self, tasks, options):
        """Creates a task in the update task queue

        :param tasks: arguments for creating tasks
        :type tasks: list.
        :param options: command line options
        :type options: list.
        """

        tpmflag = None

        path = "/redfish/v1/UpdateService/UpdateTaskQueue/"
        comps = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/" "ComponentRepository/")
        curr_tasks = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/UpdateTaskQueue/")
        for task in tasks:
            usedcomp = None
            newtask = None
            if options.targets:
                targets = []
                target_list = options.targets.split(",")
                for target in target_list:
                    target_url = "/redfish/v1/UpdateService/FirmwareInventory/" + str(target) + "/"
                    targets.append(target_url)
                verified = self.verify_targets(options.targets)
                if not verified:
                    self.rdmc.ui.error("Provided target was not available, Please provide valid target id\n")
                    return ReturnCodes.INVALID_TARGET_ERROR
            try:
                usedcomp = int(task)
                newtask = {
                    "Name": "Wait-%s %s seconds" % (str(randint(0, 1000000)), str(usedcomp)),
                    "Command": "Wait",
                    "WaitTimeSeconds": usedcomp,
                    "UpdatableBy": ["Bmc"],
                }
                if options.targets:
                    newtask["Targets"] = targets
            except ValueError:
                pass

            if task.lower() == "reboot":
                newtask = {
                    "Name": "Reboot-%s" % str(randint(0, 1000000)),
                    "Command": "ResetServer",
                    "UpdatableBy": ["RuntimeAgent"],
                }
            elif not newtask:
                if tpmflag is None:
                    if options.tover:
                        tpmflag = True
                    else:
                        tpmflag = False
                    # TODO: Update to monolith check
                    results = self.rdmc.app.get_handler(self.rdmc.app.typepath.defs.biospath, silent=True)
                    if results.status == 200:
                        contents = results.dict if self.rdmc.app.typepath.defs.isgen9 else results.dict["Attributes"]
                        tpmstate = contents["TpmState"]
                        if "Enabled" in tpmstate and not tpmflag:
                            raise IdTokenError("")

                for curr_task in curr_tasks:
                    if (
                        "Filename" in curr_task
                        and curr_task["Filename"] == task
                        and curr_task["State"].lower() != "exception"
                    ):
                        raise TaskQueueError(
                            "This file already has a task queue for flashing "
                            "associated with it. Reset the taskqueue and "
                            "retry if you need to add this task again."
                        )
                for comp in comps:
                    if comp["Filename"] == task:
                        usedcomp = comp
                        break

                if not usedcomp:
                    raise NoContentsFoundForOperationError(
                        "Component " "referenced is not present on iLO Drive: %s" % task
                    )
                newtask = {
                    "Name": "Update-%s %s"
                    % (
                        str(randint(0, 1000000)),
                        usedcomp["Name"].encode("ascii", "ignore"),
                    ),
                    "Command": "ApplyUpdate",
                    "Filename": usedcomp["Filename"],
                    "UpdatableBy": usedcomp["UpdatableBy"],
                    "TPMOverride": tpmflag,
                }
                if options.targets:
                    newtask["Targets"] = targets

            self.rdmc.ui.printer('Creating task: "%s"\n' % newtask["Name"])
            self.rdmc.ui.printer('payload: "%s"\n' % newtask)
            self.rdmc.app.post_handler(path, newtask)

    def verify_targets(self, target):
        target_list = target.split(",")
        for target in target_list:
            try:
                target_url = "/redfish/v1/UpdateService/FirmwareInventory/" + target
                dd = self.rdmc.app.get_handler(target_url, service=True, silent=True)
                if dd.status == 404:
                    return False
            except:
                return False
        return True

    def printqueue(self, options):
        """Prints the update task queue

        :param options: command line options
        :type options: list.
        """
        tasks = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/UpdateTaskQueue/")
        if not tasks:
            self.rdmc.ui.printer("No tasks found.\n")
            return

        if not options.json:
            self.rdmc.ui.printer("\nCurrent Update Task Queue:\n\n")

        if not options.json:
            for task in tasks:
                self.rdmc.ui.printer("Task %s:\n" % task["Name"])
                if "Filename" in list(task.keys()):
                    self.rdmc.ui.printer(
                        "\tCommand: %s\n\tFilename: %s\n\t"
                        "State:%s\n" % (task["Command"], task["Filename"], task["State"])
                    )
                elif "WaitTimeSeconds" in list(task.keys()):
                    self.rdmc.ui.printer(
                        "\tCommand: %s %s seconds\n\tState:%s\n"
                        % (task["Command"], str(task["WaitTimeSeconds"]), task["State"])
                    )
                else:
                    self.rdmc.ui.printer("\tCommand:%s\n\tState: %s\n" % (task["Command"], task["State"]))
        elif options.json:
            outjson = dict()
            for task in tasks:
                outjson[task["Name"]] = dict()
                outjson[task["Name"]]["Command"] = task["Command"]
                if "Filename" in task:
                    outjson[task["Name"]]["Filename"] = task["Filename"]
                if "WaitTimeSeconds" in task:
                    outjson[task["Name"]]["WaitTimeSeconds"] = task["WaitTimeSeconds"]
                outjson[task["Name"]]["State"] = task["State"]
            self.rdmc.ui.print_out_json(outjson)

    def updatetaskqueuevalidation(self, options):
        """taskqueue validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    @staticmethod
    def options_argument_group(parser):
        """Define optional arguments group

        :param parser: The parser to add the --addprivs option group to
        :type parser: ArgumentParser/OptionParser
        """
        group = parser.add_argument_group(
            "GLOBAL OPTIONS",
            "Options are available for all " "arguments within the scope of this command.",
        )

        group.add_argument(
            "--tpmover",
            dest="tover",
            action="store_true",
            help="If set then the TPMOverrideFlag is passed in on the " "associated flash operations",
            default=False,
        )

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        subcommand_parser = customparser.add_subparsers(dest="command")

        default_parser = subcommand_parser.add_parser(
            "default",
            help="Running without any sub-command will return the current task \n"
            "queue information on the currently logged in server.",
        )
        default_parser.add_argument(
            "-r",
            "--resetqueue",
            action="store_true",
            dest="resetqueue",
            help="Remove all update tasks in the queue.\n\texample: taskqueue --resetqueue or taskqueue -r",
            default=False,
        )
        default_parser.add_argument(
            "-c",
            "--cleanqueue",
            action="store_true",
            dest="cleanqueue",
            help="Clean up all finished or errored tasks left pending.\n\texample: taskqueue "
            "--cleanqueue or taskqueue -c",
            default=False,
        )
        default_parser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        self.cmdbase.add_login_arguments_group(default_parser)
        self.options_argument_group(default_parser)

        # create
        create_help = "Create a new task queue task."
        create_parser = subcommand_parser.add_parser(
            "create",
            help=create_help,
            description=create_help + "\n\n\tCreate a new task for 30 secs:\n\t\ttaskqueue "
            "create 30\n\n\tCreate a new reboot task.\n\t\ttaskqueue create reboot"
            "\n\n\tCreate a new component task.\n\t\ttaskqueue create compname"
            "\n\n\tCreate multiple tasks at once.\n\t\ttaskqueue create 30 "
            '"compname compname2 reboot"'
            "\n\n\tCreate a new task using targets: \n\t\t taskqueue create compname --targets <id>\n\n\t",
            formatter_class=RawDescriptionHelpFormatter,
        )
        create_parser.add_argument(
            "keywords",
            help="Keyword for a task queue item. *Note*: Multiple tasks can be created by "
            "using quotations wrapping all tasks, delimited by whitespace.",
            metavar="KEYWORD",
            type=str,
            nargs="+",
            default="",
        )
        create_parser.add_argument(
            "--targets",
            help="If targets value specify a comma separated\t" "firmwareinventory id only",
            metavar="targets_indices",
        )

        self.cmdbase.add_login_arguments_group(create_parser)
        self.options_argument_group(create_parser)



from __future__ import absolute_import

from telnetlib import Telnet


class IloDebugger(object):
    DEBUG_PORT = 8

    def __init__(self, hostname, timeout=10):
        """Initialize an iLO debugger instance.

        :param hostname: hostname of the iLO server
        :param timeout: connection timeout, in seconds
        """
        self.hostname = hostname
        self.timeout = timeout
        self.tn = None
        self.prompt = '\r\n>'

    def open(self):
        """Connect to the iLO debugger."""
        self.tn = Telnet(self.hostname, self.DEBUG_PORT, self.timeout)
        # wait until the debugger is ready and clear initial output
        self.read_until(self.prompt)

    def close(self):
        """Close the connection."""
        if self.tn:
            self.tn.close()
            self.tn = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def read(self):
        """Read all available output without blocking I/O."""
        output = self.tn.read_very_eager()
        # debugger occasionally outputs CRLFs with double CRs - just convert all to LFs
        output = output.replace('\r', '')
        return output

    def read_until(self, expected, timeout=10):
        """Read until an expected string is encountered or the timeout, in seconds, has been reached."""
        output = self.tn.read_until(expected, timeout)
        output = output.replace('\r', '')
        return output

    def writeline(self, line):
        """Write a line to the iLO debugger."""
        self.tn.write(line + '\n')

    def run(self, cmd_line):
        """Run an iLO debugger command and capture output."""
        self.read()  # clear any previous output
        self.writeline(cmd_line)
        output = self.read_until(self.prompt)
        return output



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" iLO Functionality Command for rdmc """

import json

try:
    from rdmc_helper import (
        IncompatableServerTypeError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IncompatableServerTypeError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )


class DisableIloFunctionalityCommand:
    """Disables iLO functionality to the server"""

    def __init__(self):
        self.ident = {
            "name": "disableilofunctionality",
            "usage": None,
            "description": "Disable iLO functionality on the current logged in server."
            "\n\texample: disableilofunctionality\n\n\tWARNING: this will"
            " render iLO unable to respond to network operations.\n\n\t"
            "Add the --force flag to ignore critical task checking.",
            "summary": "disables iLO's accessibility via the network and resets "
            "iLO. WARNING: This should be used with caution as it will "
            "render iLO unable to respond to further network operations "
            "(including REST operations) until iLO is re-enabled using the"
            " RBSU menu.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main DisableIloFunctionalityCommand function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if args:
            raise InvalidCommandLineError("disableilofunctionality command takes no arguments.")

        self.ilofunctionalityvalidation(options)

        select = "Manager."
        results = self.rdmc.app.select(selector=select)

        try:
            results = results[0]
        except:
            pass

        if results:
            path = results.resp.request.path
        else:
            raise NoContentsFoundForOperationError("Manager. not found.")

        bodydict = results.resp.dict["Oem"][self.rdmc.app.typepath.defs.oemhp]
        if bodydict["iLOFunctionalityRequired"]:
            raise IncompatableServerTypeError(
                "disableilofunctionality"
                " command is not available. iLO functionality is required"
                " and can not be disabled on this platform."
            )

        try:
            for item in bodydict["Actions"]:
                if "iLOFunctionality" in item:
                    if self.rdmc.app.typepath.defs.isgen10:
                        action = item.split("#")[-1]
                    else:
                        action = "iLOFunctionality"

                    path = bodydict["Actions"][item]["target"]
                    body = {"Action": action}
                    break
        except:
            body = {"Action": "iLOFunctionality", "Target": "/Oem/Hp"}

        if self.ilodisablechecks(options):
            self.rdmc.ui.warn(
                "Disabling iLO functionality. iLO will be unavailable on the logged "
                " in server until it is re-enabled manually.\n"
            )

            results = self.rdmc.app.post_handler(path, body, silent=True, service=True)

            if results.status == 200:
                self.rdmc.ui.printer("[%d] The operation completed successfully.\n" % results.status)
            else:
                self.rdmc.ui.printer("[%d] iLO responded with the following info: \n" % results.status)
                json_payload = json.loads(results._http_response.data)
                try:
                    self.rdmc.ui.error("%s" % json_payload["error"]["@Message.ExtendedInfo"][0]["MessageId"])
                except:
                    self.rdmc.ui.error("An invalid or incomplete response was received: %s\n" % json_payload)

        else:
            self.rdmc.ui.error(
                "iLO is currently performing a critical task and "
                "can not be safely disabled at this time. Please try again later.\n"
            )

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def ilodisablechecks(self, options):
        """Verify it is safe to actually disable iLO

        :param options: command line options
        :type options: values, attributes of class obj
        """

        if options.force:
            self.rdmc.ui.warn("Force Enabled: Ignoring critical operation/mode checking.\n")
            return True

        else:
            keyword_list = "idle", "complete"

            try:
                results = self.rdmc.app.select(selector="UpdateService.")[0]
            except:
                raise NoContentsFoundForOperationError("UpdateService. not found.")

            try:
                state = results.resp.dict["Oem"]["Hpe"]["State"].lower()
                for val in keyword_list:
                    if val in state:
                        return True
                return False

            except:
                raise NoContentsFoundForOperationError("iLO state not identified")

    def ilofunctionalityvalidation(self, options):
        """ilofunctionalityvalidation method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--force",
            dest="force",
            help="Ignore any critical task checking and force disable iLO.",
            action="store_true",
            default=None,
        )



# ##
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ##

# -*- coding: utf-8 -*-
""" Update Task Queue Command for rdmc """

import re
from argparse import RawDescriptionHelpFormatter
from random import randint

from redfish.ris.rmc_helper import ValidationError

try:
    from rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IncompatibleiLOVersionError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )

__subparsers__ = ["add", "delete"]


class MaintenanceWindowCommand:
    """Main maintenancewindow command class"""

    def __init__(self):
        self.ident = {
            "name": "maintenancewindow",
            "usage": None,
            "description": "Add or delete maintenance windows from the iLO repository.\nTo "
            " view help on specific sub-commands run: maintenancewindow <sub-command> -h\n\n"
            "NOTE: iLO 5 required.",
            "summary": "Manages the maintenance windows for iLO.",
            "aliases": [],
            "auxcommands": [],
        }

    def run(self, line, help_disp=False):
        """Main update maintenance window worker function

        :param line: string of arguments passed in
        :type line: str.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            ident_subparser = False
            for cmnd in __subparsers__:
                if cmnd in line:
                    (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
                    ident_subparser = True
                    break
            if not ident_subparser:
                (options, args) = self.rdmc.rdmc_parse_arglist(self, line, default=True)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.maintenancewindowvalidation(options)

        if self.rdmc.app.typepath.defs.isgen9:
            raise IncompatibleiLOVersionError("iLO Repository commands are only available on iLO 5.")

        windows = self.rdmc.app.getcollectionmembers("/redfish/v1/UpdateService/MaintenanceWindows/")

        if options.command.lower() == "add":
            self.addmaintenancewindow(options, windows, options.time_window)
        elif options.command.lower() == "delete":
            self.deletemaintenancewindow(windows, options.identifier)
        else:
            self.listmainenancewindows(options, windows)

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def addmaintenancewindow(self, options, windows, startafter):
        """Add a maintenance window

        :param options: command line options
        :type options: list.
        :param windows: list of maintenance windows on the system
        :type windows: list.
        :param startafter: redfish date-time string to start the maintenance window
        :type startafter: str.
        """
        adddata = {"StartAfter": startafter}

        if options.name:
            adddata["Name"] = options.name
        else:
            adddata["Name"] = "MW-%s" % str(randint(0, 1000000))

        if options.description:
            if options.description.startswith('"') and options.description.endswith('"'):
                options.description = options.description[1:-1]
            adddata["Description"] = options.description

        if options.expire:
            adddata["Expire"] = options.expire

        errors = self.validatewindow(adddata, windows)

        if not errors:
            path = "/redfish/v1/UpdateService/MaintenanceWindows/"
            self.rdmc.app.post_handler(path, adddata)
        else:
            self.rdmc.ui.error("Invalid Maintenance Window:\n")
            for error in errors:
                self.rdmc.ui.error("\t" + error + "\n")
            raise ValidationError("")

    def deletemaintenancewindow(self, windows, nameid):
        """Delete a maintenance window by Id or Name

        :param windows: list of maintenance windows on the system
        :type windows: list.
        :param nameid: id or name string to remove
        :type nameid: str.
        """

        deleted = False
        for window in windows:
            if window["Name"] == nameid or window["Id"] == nameid:
                path = window["@odata.id"]
                self.rdmc.ui.printer("Deleting %s\n" % nameid)
                self.rdmc.app.delete_handler(path)
                deleted = True
                break

        if not deleted:
            raise NoContentsFoundForOperationError("No maintenance window found with that Name/Id.")

    def listmainenancewindows(self, options, windows):
        """Lists the maintenance windows

        :param options: command line options
        :type options: list.
        :param windows: list of maintenance windows on the system
        :type windows: list.
        """

        outstring = ""
        jsonwindows = []

        if windows:
            for window in windows:
                if options.json:
                    jsonwindows.append(dict((key, val) for key, val in window.items() if "@odata." not in key))
                else:
                    outstring += "%s:" % window["Name"]
                    outstring += "%s:" % window["Id"]
                    if "Description" in list(window.keys()) and window["Description"]:
                        outstring += "\n\tDescription: %s" % window["Description"]
                    else:
                        outstring += "\n\tDescription: %s" % "No description."
                    outstring += "\n\tStart After: %s" % window["StartAfter"]
                    if "Expire" in list(window.keys()):
                        outstring += "\n\tExpires at: %s" % window["Expire"]
                    else:
                        outstring += "\n\tExpires at: %s" % "No expire time set."
                    outstring += "\n"
            if jsonwindows:
                self.rdmc.ui.print_out_json(jsonwindows)
            else:
                self.rdmc.ui.printer(outstring)
        else:
            self.rdmc.ui.warn("No maintenance windows found on system.\n")

    def validatewindow(self, cmw, windows):
        """Validate the maintenance window before adding

        :param cmw: a maintenance window candidate
        :type cmw: dict.
        :param windows: list of maintenance windows on the system
        :type windows: list.
        :returns: returns a list of errors or a empty list if no errors
        """
        errorlist = []
        rfdtregex = "\\d\\d\\d\\d-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\dZ?"

        for window in windows:
            if cmw["Name"] == window["Name"]:
                errorlist.append("Maintenance window with Name: %s already exists." % cmw["Name"])

        if "Name" in list(cmw.keys()):
            if len(cmw["Name"]) > 64:
                errorlist.append("Name must be 64 characters or less.")
        if "Description" in list(cmw.keys()):
            if len(cmw["Description"]) > 64:
                errorlist.append("Description must be 64 characters or less.")
        if "Expire" in list(cmw.keys()):
            if not re.match(rfdtregex, cmw["Expire"]):
                errorlist.append(
                    "Invalid redfish date-time format in Expire. "
                    "Accepted formats: YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ssZ"
                )
        if not re.match(rfdtregex, cmw["StartAfter"]):
            errorlist.append(
                "Invalid redfish date-time format in StartAfter. "
                "Accepted formats YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ssZ"
            )

        return errorlist

    def maintenancewindowvalidation(self, options):
        """maintenencewindow validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        subcommand_parser = customparser.add_subparsers(dest="command")

        # default sub-parser
        default_parser = subcommand_parser.add_parser(
            "default",
            help="Running without any sub-command will return all maintenace windows on the "
            "currently logged in server.",
        )
        default_parser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        self.cmdbase.add_login_arguments_group(default_parser)

        # add sub-parser
        add_help = "Adds a maintenance window to iLO"
        add_parser = subcommand_parser.add_parser(
            "add",
            help=add_help,
            description=add_help + "\nexample: maintenancewindow add 1998-11-21T00:00:00 "
            "--expire=1998-11-22T00:00:00 --name=MyMaintenanceWindow --description "
            '"My maintenance window description.,"',
            formatter_class=RawDescriptionHelpFormatter,
        )
        add_parser.add_argument(
            "time_window",
            help="Specify the time window start period in DateTime format.\ni.e. YEAR-MONTH-DAY"
            "THOUR:MINUTE:SECOND.\nexample: 1998-11-21T10:59:58",
            metavar="TIMEWINDOW",
        )
        add_parser.add_argument(
            "--description",
            dest="description",
            help="Optionally include this flag if you would like to add a "
            "description to the maintenance window you create",
            default=None,
        )
        add_parser.add_argument(
            "--name",
            dest="name",
            help="Optionally include this flag if you would like to add a "
            "name to the maintenance window you create. If you do not specify one"
            " a unique name will be added.",
            default=None,
        )
        add_parser.add_argument(
            "--expire",
            dest="expire",
            help="Optionally include this flag if you would like to add a " "time the maintenance window expires.",
            default=None,
        )
        self.cmdbase.add_login_arguments_group(add_parser)

        # delete sub-parser
        delete_help = "Deletes the specified maintenance window on the currently logged in server."
        delete_parser = subcommand_parser.add_parser(
            "delete",
            help=delete_help,
            description=delete_help + "\nexample: maintenancewindow delete mymaintenancewindowname\n"
            "Note: The maintenance window identifier can be referenced by Name or ID#."
            "maintenancewindow delete name",
            formatter_class=RawDescriptionHelpFormatter,
        )
        delete_parser.add_argument(
            "identifier",
            help="The unique identifier provided by iLO or the identifier provided by '--name' "
            "when the maintenance window was created.",
        )
        self.cmdbase.add_login_arguments_group(delete_parser)



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Results Command for rdmc """

from redfish.ris.resp_handler import ResponseHandler
from redfish.ris.rmc_helper import EmptyRaiseForEAFP

try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )


class ResultsCommand:
    """Monolith class command"""

    def __init__(self):
        self.ident = {
            "name": "results",
            "usage": None,
            "description": "Run to show the results of the last" " changes after a server reboot.\n\texample: results",
            "summary": "Show the results of changes which require a server reboot.",
            "aliases": [],
            "auxcommands": ["LoginCommand", "SelectCommand"],
        }

        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Gather results of latest BIOS change

        :param line: string of arguments passed in
        :type line: str.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if args:
            raise InvalidCommandLineError("Results command does not take any arguments.")
        self.resultsvalidation(options)
        results = {}
        if self.rdmc.app.typepath.defs.biospath[-1] == "/":
            iscsipath = self.rdmc.app.typepath.defs.biospath + "iScsi/"
            bootpath = self.rdmc.app.typepath.defs.biospath + "Boot/"
        else:
            iscsipath = self.rdmc.app.typepath.defs.biospath + "/iScsi"
            bootpath = self.rdmc.app.typepath.defs.biospath + "/Boot"

        try:
            self.auxcommands["select"].selectfunction("SmartStorageConfig")
            smartarray = self.rdmc.app.getprops()
            sapaths = [path["@odata.id"].split("settings")[0] for path in smartarray]
        except:
            sapaths = None

        biosresults = self.rdmc.app.get_handler(self.rdmc.app.typepath.defs.biospath, service=True, silent=True)
        iscsiresults = self.rdmc.app.get_handler(iscsipath, service=True, silent=True)
        bootsresults = self.rdmc.app.get_handler(bootpath, service=True, silent=True)
        saresults = []
        if sapaths:
            saresults = [self.rdmc.app.get_handler(path, service=True, silent=True) for path in sapaths]
        try:
            results.update({"Bios:": biosresults.dict[self.rdmc.app.typepath.defs.biossettingsstring]["Messages"]})
        except Exception:
            results.update({"Bios:": None})

        try:
            results.update({"Iscsi:": iscsiresults.dict[self.rdmc.app.typepath.defs.biossettingsstring]["Messages"]})
        except:
            results.update({"Iscsi:": None})

        try:
            results.update({"Boot:": bootsresults.dict[self.rdmc.app.typepath.defs.biossettingsstring]["Messages"]})
        except:
            results.update({"Boot:": None})
        try:
            for result in saresults:
                loc = "SmartArray"
                if saresults.index(result) > 0:
                    loc += " %d:" % saresults.index(result)
                else:
                    loc += ":"
                results.update({loc: result.dict[self.rdmc.app.typepath.defs.biossettingsstring]["Messages"]})
        except:
            results.update({"SmartArray:": None})

        self.rdmc.ui.printer("Results of the previous reboot changes:\n\n")

        for result in results:
            self.rdmc.ui.printer("%s\n" % result)
            try:
                for msg in results[result]:
                    _ = ResponseHandler(
                        self.rdmc.app.validationmanager,
                        self.rdmc.app.typepath.defs.messageregistrytype,
                    ).message_handler(response_data=msg, message_text="", verbosity=0, dl_reg=False)
                    pass
            except EmptyRaiseForEAFP as exp:
                raise EmptyRaiseForEAFP(exp)
            except Exception:
                self.rdmc.ui.error("No messages found for %s.\n" % result[:-1])

        self.cmdbase.logout_routine(self, options)
        return ReturnCodes.SUCCESS

    def resultsvalidation(self, options):
        """Results method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)



"""
conftest.py is used to set up shared fixtures, plugins, hooks, other config etc.
This is the global, root-level conftest.py loaded for all tests.
Sub-directories can have their own conftest.py that gets loaded locally for that sub-directory only.
http://docs.pytest.org/en/latest/writing_plugins.html#conftest-py-local-per-directory-plugins
"""
from __future__ import absolute_import

import os
import sys
from importlib import import_module

import pytest

# Add the 'src' directory to the module search path so our tests can import from it
SRC_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../ilorest'))
RDMC_FILE = os.path.join(SRC_DIR, 'rdmc.py')

sys.path.insert(0, SRC_DIR)

DisplayHelpers = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.lib.DisplayHelpers')
ShowPmemCommand = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.ShowPmemCommand')
RestHelpers = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.lib.RestHelpers')
Mapper = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.lib.Mapper')
MapperRenderers = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.lib.MapperRenderers')
PmemHelpers = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.lib.PmemHelpers')
ClearPendingConfigCommand = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.ClearPendingConfigCommand')
ApplyPmemConfigCommand = import_module('extensions.PERSISTENT_MEMORY_COMMANDS.ApplyPmemConfigCommand')

def pytest_addoption(parser):
    parser.addoption("--runquix", action="store_true", default=False,
                     help="Run tests associated with an open QuIX (which are skipped by default).")
    parser.addoption("--skipincremental", action="store_true", default=False,
                     help="Skip incremental configuration tests (run by default).")
    parser.addoption("--clear-nand-counters", action="store_true", default=False,
                     help="Clear all iLO NAND write counters using the iLO debugger.")

    logging_group = parser.getgroup('logging', description='logging configuration')
    logging_group.addoption("--ilodebug", metavar='path', help="log iLO debugger output to the specified file.")
    logging_group.addoption("--vsp", metavar='path', help="log Virtual Serial Port output to the specified file.")
    logging_group.addoption("--uefi-debug-level", choices=[0, 1, 2, 3], type=int,
                            help="set UEFI serial debug level when capturing Virtual Serial Port output: "
                                 "0 (Disabled), 1 (ErrorsOnly), 2 (Medium), 3 (Verbose)")


def pytest_runtest_setup(item):
    openquix_marker = item.get_marker("openquix")
    if openquix_marker is not None:
        if item.config.getoption("--runquix") is False:
            pytest.skip("skipping test associated with an open QuIX")


def pytest_collection(session):
    """
    Hook to use testpaths for a test root directory passed in as a pytest argument.
    This allows using testpaths without having to execute pytest from the test root directory.
    """
    args = session.config.args
    testpaths = session.config.getini('testpaths')
    currdir = session.startdir
    rootdir = session.config.rootdir

    if len(args) == 1 and testpaths:
        testdir = args[0]
        if currdir.join(testdir) == rootdir:
            session.config.args = [os.path.join(testdir, testpath) for testpath in testpaths]



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Types Command for RDMC """

try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )


class TypesCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "types",
            "usage": None,
            "description": "Run to display currently " "available selectable types\n\tExample: types",
            "summary": "Displays all selectable types within the currently logged in server.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def typesfunction(self, line, returntypes=False):
        """Main types worker function

        :param line: command line input
        :type line: string.
        :param returntypes: flag to determine if types should be printed
        :type returntypes: boolean.
        """
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.typesvalidation(options)

        if not args:
            typeslist = list()
            typeslist = sorted(set(self.rdmc.app.types(options.fulltypes)))

            if not returntypes:
                self.rdmc.ui.printer("Type options:\n")

                for item in typeslist:
                    self.rdmc.ui.printer("%s\n" % item)
            else:
                return typeslist
        else:
            raise InvalidCommandLineError("The 'types' command does not take any arguments.")

        self.cmdbase.logout_routine(self, options)

    def run(self, line, help_disp=False):
        """Wrapper function for types main function

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        self.typesfunction(line)

        # Return code
        return ReturnCodes.SUCCESS

    def typesvalidation(self, options):
        """types method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--fulltypes",
            dest="fulltypes",
            action="store_true",
            help="Optionally include this flag if you would prefer to "
            "return the full type name instead of the simplified versions"
            " (Redfish only option).",
            default=None,
        )
